# Copyright 2019-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os

import pytest
import yaml

from klio_core import config

from klio.metrics import logger as logger_metrics
from klio.metrics import stackdriver as sd_metrics
from klio.transforms import core as core_transforms


@pytest.mark.parametrize(
    "runner,metrics_config,exp_clients,exp_logger_enabled",
    (
        # default metrics client config for respective runner
        ("dataflow", {}, (sd_metrics.StackdriverLogMetricsClient,), None),
        ("direct", {}, (logger_metrics.MetricsLoggerClient,), True),
        # explicitly turn on metrics for respective runner
        (
            "dataflow",
            {"stackdriver_logger": True},
            (sd_metrics.StackdriverLogMetricsClient,),
            None,
        ),
        (
            "direct",
            {"logger": True},
            (logger_metrics.MetricsLoggerClient,),
            True,
        ),
        # turn off default client for respective runner
        (
            "dataflow",
            {"stackdriver_logger": False},
            (logger_metrics.MetricsLoggerClient,),
            False,
        ),
        (
            "direct",
            {"logger": False},
            (logger_metrics.MetricsLoggerClient,),
            False,
        ),
        # ignore SD config when on direct
        (
            "direct",
            {"stackdriver_logger": True},
            (logger_metrics.MetricsLoggerClient,),
            True,
        ),
        # ignore logger config when on dataflow
        (
            "dataflow",
            {"logger": False},
            (sd_metrics.StackdriverLogMetricsClient,),
            None,
        ),
    ),
)
def test_klio_metrics(
    runner,
    metrics_config,
    exp_clients,
    exp_logger_enabled,
    klio_config,
    mocker,
    monkeypatch,
):
    klio_ns = core_transforms.KlioContext()
    # sanity check / clear out thread local
    klio_ns._thread_local.klio_metrics = None

    monkeypatch.setattr(klio_config.pipeline_options, "runner", runner, None)
    monkeypatch.setattr(klio_config.job_config, "metrics", metrics_config)
    mock_config = mocker.PropertyMock(return_value=klio_config)
    monkeypatch.setattr(core_transforms.KlioContext, "config", mock_config)

    registry = klio_ns.metrics

    for actual_relay in registry._relays:
        assert isinstance(actual_relay, exp_clients)
        if isinstance(actual_relay, logger_metrics.MetricsLoggerClient):
            assert exp_logger_enabled is not actual_relay.disabled


@pytest.mark.parametrize("usr_local_exists", (True, False))
@pytest.mark.parametrize("usr_glob_exists", (True, False))
def test_load_config_from_file(
    usr_local_exists, usr_glob_exists, config_dict, mocker, monkeypatch,
):
    monkeypatch.setattr(os.path, "exists", lambda x: usr_local_exists)
    effective_klio_yaml_file = "/usr/src/app/klio-job-run-effective.yaml"
    expected_open_file = "/usr/src/config/.effective-klio-job.yaml"

    if usr_local_exists:
        expected_open_file = "/usr/local/klio-job-run-effective.yaml"
    elif usr_glob_exists:
        expected_open_file = effective_klio_yaml_file

    mock_iglob = mocker.Mock()
    if usr_glob_exists:
        mock_iglob.return_value = [effective_klio_yaml_file]
    else:
        mock_iglob.return_value = []

    monkeypatch.setattr(core_transforms.glob, "iglob", mock_iglob)

    open_name = "klio.transforms.core.open"
    config_str = yaml.dump(config_dict)
    m_open = mocker.mock_open(read_data=config_str)
    m = mocker.patch(open_name, m_open)

    klio_config = core_transforms.RunConfig._load_config_from_file()

    m.assert_called_once_with(expected_open_file, "r")
    assert isinstance(klio_config, config.KlioConfig)
    if not usr_local_exists:
        mock_iglob.assert_called_once_with(
            "/usr/**/klio-job-run-effective.yaml", recursive=True
        )
    else:
        mock_iglob.assert_not_called()


def test_load_config_from_file_raises(config_dict, mocker, monkeypatch):
    monkeypatch.setattr(os.path, "exists", lambda x: False)

    mock_iglob = mocker.Mock()
    mock_iglob.return_value = iter([])
    monkeypatch.setattr(core_transforms.glob, "iglob", mock_iglob)

    with pytest.raises(IOError):
        core_transforms.RunConfig._load_config_from_file()


@pytest.mark.parametrize("thread_local_ret", (True, False))
def test_job_property(thread_local_ret, mocker, monkeypatch):
    mock_func = mocker.Mock()
    monkeypatch.setattr(
        core_transforms.KlioContext, "_create_klio_job_obj", mock_func
    )

    klio_ns = core_transforms.KlioContext()

    if not thread_local_ret:
        # sanity check / clear out thread local
        klio_ns._thread_local.klio_job = None
    else:
        klio_ns._thread_local.klio_job = mocker.Mock()

    ret_value = klio_ns.job

    if not thread_local_ret:
        mock_func.assert_called_once_with()
        assert (
            mock_func.return_value
            == ret_value
            == klio_ns._thread_local.klio_job
        )
    else:
        mock_func.assert_not_called()
        assert klio_ns._thread_local.klio_job == ret_value

    klio_ns._thread_local.klio_job = None


@pytest.mark.parametrize("thread_local_ret", (True, False))
def test_metrics_property(thread_local_ret, mocker, monkeypatch):
    mock_func = mocker.Mock()
    monkeypatch.setattr(
        core_transforms.KlioContext, "_get_metrics_registry", mock_func
    )

    klio_ns = core_transforms.KlioContext()

    if not thread_local_ret:
        # sanity check / clear out thread local
        klio_ns._thread_local.klio_metrics = None
    else:
        klio_ns._thread_local.klio_metrics = mocker.Mock()

    ret_value = klio_ns.metrics

    if not thread_local_ret:
        mock_func.assert_called_once_with()
        assert (
            mock_func.return_value
            == ret_value
            == klio_ns._thread_local.klio_metrics
        )
    else:
        mock_func.assert_not_called()
        assert klio_ns._thread_local.klio_metrics == ret_value



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Collection of commonly used transforms and transform-utilities."""

from klio.transforms.io import (
    KlioReadFromAvro,
    KlioReadFromBigQuery,
    KlioReadFromText,
    KlioWriteToAvro,
    KlioWriteToBigQuery,
    KlioWriteToText,
)


__all__ = (
    "KlioReadFromAvro",
    "KlioReadFromBigQuery",
    "KlioReadFromText",
    "KlioWriteToAvro",
    "KlioWriteToBigQuery",
    "KlioWriteToText",
)



# Copyright 2019-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


class KlioMessageException(Exception):
    """General :ref:`KlioMessage <klio-message>` exception."""


class KlioMessagePayloadException(KlioMessageException):
    """Error handling payload for a :ref:`KlioMessage <klio-message>`."""



# Copyright 2019-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__author__ = "The klio developers"
__version__ = "0.2.5rc3"
__email__ = "opensource+klio@spotify.com"
__description__ = "Conventions for Python + Apache Beam "
__uri__ = "https://github.com/spotify/klio"



# Copyright 2019-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging

import pytest

from klio.metrics import dispatcher


class GenericDispatcher(dispatcher.BaseMetricDispatcher):

    METRIC_TYPE = "generic"

    def _setup_metric_relay(self, relay_clients):
        pass


@pytest.fixture
def metric_params():
    return {
        "name": "my-metric",
        "value": 0,
        "transform": "my-transform",
        "kwargs": {"tags": {"key-tag": "value-tag"}},
    }


@pytest.fixture
def metric(mocker):
    mock = mocker.Mock()
    mock.update = mocker.Mock()
    return mock


@pytest.fixture
def relay_client(mocker, metric):
    client = mocker.Mock()
    client.emit = mocker.Mock()
    client.counter.return_value = metric
    client.gauge.return_value = metric
    client.timer.return_value = metric
    return client


@pytest.fixture
def counter_dispatcher(
    relay_client, metric_params, metric, mocker, monkeypatch
):
    counter = dispatcher.CounterDispatcher(
        relay_clients=[relay_client], **metric_params
    )
    mock_submit_to_thread = mocker.Mock()
    monkeypatch.setattr(counter, "submit", mock_submit_to_thread)
    return counter


@pytest.fixture
def gauge_dispatcher(relay_client, metric_params, metric, mocker, monkeypatch):
    gauge = dispatcher.GaugeDispatcher(
        relay_clients=[relay_client], **metric_params
    )
    mock_submit_to_thread = mocker.Mock()
    monkeypatch.setattr(gauge, "submit", mock_submit_to_thread)
    return gauge


@pytest.fixture
def default_timer(mocker, monkeypatch):
    mock = mocker.Mock()
    mock.default_timer.side_effect = [0.0, 1.0]
    monkeypatch.setattr(dispatcher, "timeit", mock)
    return mock


@pytest.fixture
def timer_dispatcher(
    relay_client, metric_params, metric, default_timer, mocker, monkeypatch
):
    timer = dispatcher.TimerDispatcher(
        relay_clients=[relay_client], **metric_params
    )
    mock_submit_to_thread = mocker.Mock()
    monkeypatch.setattr(timer, "submit", mock_submit_to_thread)
    return timer


@pytest.fixture
def generic_dispatcher(relay_client, metric_params):
    return GenericDispatcher(relay_clients=[relay_client], **metric_params)


def test_base_metric_dispatcher_raises(relay_client, metric_params):
    class FakeMetricDispatcher(dispatcher.BaseMetricDispatcher):
        pass

    with pytest.raises(NotImplementedError):
        FakeMetricDispatcher(relay_clients=[relay_client], **metric_params)


@pytest.mark.parametrize(
    "has_transform,exp_key",
    ((True, "generic_my-metric_my-transform"), (False, "generic_my-metric")),
)
def test_base_metric_setup_metric_key(
    has_transform, exp_key, relay_client, metric_params
):
    if not has_transform:
        metric_params.pop("transform")

    gd = GenericDispatcher(relay_clients=[relay_client], **metric_params)
    assert exp_key == gd.metric_key


def test_base_metric_logger(generic_dispatcher):
    assert not getattr(
        generic_dispatcher._thread_local,
        "klio_metrics_dispatcher_logger",
        None,
    )  # sanity check

    logger = generic_dispatcher.logger
    assert logging.getLogger("klio.metrics.dispatcher") == logger
    assert (
        logger
        == generic_dispatcher._thread_local.klio_metrics_dispatcher_logger
    )


def test_base_metric_submit_callback(generic_dispatcher, mocker, caplog):
    fut = mocker.Mock()
    fut.result.return_value = None

    generic_dispatcher._submit_callback(fut)

    fut.result.assert_called_once_with()
    assert not len(caplog.records)


def test_base_metric_submit_callback_exception(
    generic_dispatcher, mocker, caplog
):
    fut = mocker.Mock()
    fut.result.side_effect = Exception("fuu")

    generic_dispatcher._submit_callback(fut)

    fut.result.assert_called_once_with()
    assert 1 == len(caplog.records)
    assert logging.WARNING == caplog.records[0].levelno


def test_base_metric_submit(
    generic_dispatcher, relay_client, mocker, monkeypatch
):
    mock_thread_pool, mock_future = mocker.Mock(), mocker.Mock()
    mock_thread_pool.submit.return_value = mock_future
    monkeypatch.setattr(generic_dispatcher, "_thread_pool", mock_thread_pool)

    metric = mocker.Mock()
    metric.name, metric.transform = "my-metric", "my-transform"
    expected_key = "generic_my-metric_my-transform"

    generic_dispatcher.submit(relay_client.emit, metric)

    mock_thread_pool.submit.assert_called_once_with(relay_client.emit, metric)
    assert expected_key == mock_future.metric_key
    mock_future.add_done_callback.assert_called_once_with(
        generic_dispatcher._submit_callback
    )


def test_counter_setup_metric_relay(relay_client, metric, metric_params):
    counter = dispatcher.CounterDispatcher(
        relay_clients=[relay_client], **metric_params
    )

    expected_relay_to_metric = [(relay_client, metric)]
    assert expected_relay_to_metric == counter.relay_to_metric
    relay_client.counter.assert_called_once_with(**metric_params)


def test_counter_inc(counter_dispatcher, relay_client, metric):
    assert 0 == counter_dispatcher.value  # sanity check

    counter_dispatcher.inc()

    assert 1 == counter_dispatcher.value
    metric.update.assert_called_once_with(1)
    counter_dispatcher.submit.assert_called_once_with(
        relay_client.emit, metric
    )


def test_gauge_setup_metric_relay(relay_client, metric, metric_params):
    gauge = dispatcher.GaugeDispatcher(
        relay_clients=[relay_client], **metric_params
    )

    expected_relay_to_metric = [(relay_client, metric)]
    assert expected_relay_to_metric == gauge.relay_to_metric
    relay_client.gauge.assert_called_once_with(**metric_params)


def test_gauge_set(gauge_dispatcher, relay_client, metric):
    assert 0 == gauge_dispatcher.value  # sanity check

    gauge_dispatcher.set(5)

    assert 5 == gauge_dispatcher.value
    metric.update.assert_called_once_with(5)
    gauge_dispatcher.submit.assert_called_once_with(relay_client.emit, metric)


def test_timer_setup_metric_relay(relay_client, metric, metric_params):
    metric_params["timer_unit"] = "ns"
    timer = dispatcher.TimerDispatcher(
        relay_clients=[relay_client], **metric_params
    )

    expected_relay_to_metric = [(relay_client, metric)]
    assert expected_relay_to_metric == timer.relay_to_metric
    relay_client.timer.assert_called_once_with(**metric_params)


def test_timer_start(timer_dispatcher, metric):
    assert 0 == timer_dispatcher.value  # sanity check
    assert not timer_dispatcher._start_time  # sanity check

    timer_dispatcher.start()

    assert 0.0 == timer_dispatcher._start_time
    metric.update.assert_not_called()
    timer_dispatcher.submit.assert_not_called()


def test_timer_stop(timer_dispatcher, relay_client, metric):
    assert 0 == timer_dispatcher.value  # sanity check
    assert not timer_dispatcher._start_time  # sanity check

    timer_dispatcher.start()

    assert 0.0 == timer_dispatcher._start_time
    metric.update.assert_not_called()
    timer_dispatcher.submit.assert_not_called()

    timer_dispatcher.stop()
    # seconds -> nanoseconds
    assert 1.0 * 1e9 == timer_dispatcher.value
    metric.update.assert_called_once_with(timer_dispatcher.value)
    timer_dispatcher.submit.assert_called_once_with(relay_client.emit, metric)


def test_timer_stop_no_start(timer_dispatcher, metric, caplog):
    assert not timer_dispatcher._start_time  # sanity check

    timer_dispatcher.stop()

    assert 1 == len(caplog.records)
    assert logging.WARNING == caplog.records[0].levelno

    metric.update.assert_not_called()
    timer_dispatcher.submit.assert_not_called()


def test_timer_context_manager(
    relay_client, metric, metric_params, default_timer, mocker, monkeypatch
):
    timer = dispatcher.TimerDispatcher(
        relay_clients=[relay_client], **metric_params
    )
    mock_submit_to_thread = mocker.Mock()
    monkeypatch.setattr(timer, "submit", mock_submit_to_thread)
    with timer:
        assert 1 == 1

    assert 0 == timer._start_time
    assert 1 * 1e9 == timer.value
    metric.update.assert_called_once_with(timer.value)
    timer.submit.assert_called_once_with(relay_client.emit, metric)



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module is inspired by https://github.com/invl/retry
and https://github.com/pnpnpn/retry-decorator.
It's been simplified for what we need, and adapted to work with generators.
"""

import logging
import threading
import types


TRACEBACK_HEADER = "Traceback (most recent call last):\n"


class KlioRetriesExhausted(Exception):
    """No more retries left."""


class KlioRetryWrapper(object):
    """Wrap a function with configured retries."""

    DEFAULT_EXC_MSG = (
        "Function '{}' has reached the maximum {} retries. Last exception: {}"
    )

    def __init__(
        self,
        function,
        tries,
        delay=None,
        exception=None,
        raise_exception=None,
        exception_message=None,
    ):
        self._function = function
        self._func_name = getattr(function, "__qualname__", function.__name__)
        self._tries = tries
        self._delay = delay or 0.0
        self._exception = exception or Exception
        self._retry_exception = raise_exception or KlioRetriesExhausted
        self._exception_message = exception_message
        self._logger = logging.getLogger("klio")

    def __call__(self, *args, **kwargs):
        tries = self._tries

        while tries:
            try:
                ret = self._function(*args, **kwargs)
                if isinstance(ret, types.GeneratorType):
                    ret = next(ret)
                return ret

            except self._exception as e:
                tries -= 1
                if not tries:
                    self._raise_exception(e)
                    break

                msg = self._format_log_message(tries, e)
                self._logger.warning(msg)

                if not self._delay > 0.0:
                    continue

                # use a Condition to avoid using `time.sleep` as that will
                # block other threads; we just want to block this thread
                condition = threading.Condition()
                condition.acquire()
                condition.wait(self._delay)
                condition.release()

    def _format_log_message(self, tries, exception):
        msg_prefix = "Retrying KlioMessage"
        if self._delay > 0:
            msg_prefix = "{} in {} seconds".format(msg_prefix, self._delay)

        attempts = self._tries - tries
        msg_attempts = "attempt {}".format(attempts)
        if self._tries > 0:
            msg_attempts = "{} of {}".format(msg_attempts, self._tries)

        msg = "{} - {}. '{}' raised an exception: {} ".format(
            msg_prefix, msg_attempts, self._func_name, exception
        )

        return msg

    def _raise_exception(self, last_exception):
        if self._exception_message is None:
            self._exception_message = self.DEFAULT_EXC_MSG.format(
                self._func_name, self._tries, last_exception
            )
        error = self._retry_exception(self._exception_message)
        self._logger.error(
            "Dropping KlioMessage - retries exhausted: {}".format(error)
        )
        raise error



# Copyright 2019-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Klio ships with a Stackdriver Log-based Metrics relay client. The client
creates metrics objects in Stackdriver Monitoring based off of filters in
Stackdriver Logging. For more information on Stackdriver's Log-based Metrics,
see `related documentation <https://cloud.google.com/logging/docs/
logs-based-metrics/>`_.

When running on Dataflow, the client is on by default with no additional
configuration needed. This must be actively turned off in ``klio-job.yaml``
if not wanted (see below).

The log-based metrics client is not available for direct runner.

To explicitly turn off log-based metrics, in ``klio-job.yaml``:

.. code-block:: yaml

    job_config:
      metrics:
        stackdriver_logger: false

"""
import logging

from googleapiclient import discovery
from googleapiclient import errors as gapi_errors

from klio.metrics import logger


class StackdriverLogMetricsClient(logger.MetricsLoggerClient):
    """Stackdriver client for transform metrics.

    Intended to be instantiated by :class:`klio.metrics.client.MetricsRegistry`
    and not by itself.

    Args:
        klio_config (klio_core.config.KlioConfig):  the job's configuration.
    """

    RELAY_CLIENT_NAME = "stackdriver_logger"

    def __init__(self, klio_config):
        super(StackdriverLogMetricsClient, self).__init__(klio_config)
        self.job_name = klio_config.job_name

    @property
    def _stackdriver_client(self):
        client = getattr(self._thread_local, "stackdriver_client", None)
        if not client:
            # FYI this does not make a network call
            client = discovery.build("logging", "v2")
            self._thread_local.stackdriver_client = client
        return self._thread_local.stackdriver_client

    # get a list of metrics first and use any already-available metrics
    # to create metrics objects
    def counter(self, name, transform=None, tags=None, **kwargs):
        """Create a :class:`StackdriverLogMetricsCounter` object.

        .. note::

            Stackdriver counts log lines so initializing a
            counter value is not supported .

        Args:
            name (str): name of counter
            transform (str): transform the counter is associated with
            tags (dict): any tags of additional contextual information
                to associate with the counter

        Returns:
            StackdriverLogMetricsCounter: a log-based counter
        """
        # Since stackdriver literally counts loglines, initializing a
        # counter value is not supported
        if kwargs.get("value", 0) > 0:
            self.logger.log(
                logging.WARNING,
                "Initializing Stackdriver log-based counters with a value "
                "other than 0 is not supported. Defaulting to 0.",
            )
        ctr = StackdriverLogMetricsCounter(
            name=name,
            job_name=self.job_name,
            project=self.klio_config.pipeline_options.project,
            transform=transform,
            tags=tags,
        )
        ctr._init_metric(self._stackdriver_client)
        return ctr

    def gauge(self, *args, **kwargs):
        """Create a :class:`StackdriverLogMetricsGauge` object.

        .. warning::

            Gauges for Stackdriver are not yet supported. This will
            default to standard logging.

        Args:
            name (str): name of gauge
            value (int): starting value of gauge; defaults to 0
            transform (str): transform the gauge is associated with
            tags (dict): any tags of additional contextual information
                to associate with the gauge

        Returns:
            StackdriverLogMetricsGauge: a log-based gauge
        """
        self.logger.log(
            logging.WARNING,
            "Gauge is not supported for Stackdriver log-based metrics, "
            "defaulting to standard logger.",
        )
        return StackdriverLogMetricsGauge(*args, **kwargs)

    def timer(self, *args, **kwargs):
        """Create a :class:`StackdriverLogMetricsTimer` object.

        .. warning::

            Timers for Stackdriver are not yet supported. This will
            default to standard logging.

        Args:
            name (str): name of gauge
            value (int): starting value of gauge; defaults to 0
            transform (str): transform the gauge is associated with
            tags (dict): any tags of additional contextual information
                to associate with the gauge

        Returns:
            StackdriverLogMetricsTimer: a log-based timer
        """
        self.logger.log(
            logging.WARNING,
            "Timer is not supported for Stackdriver log-based metrics, "
            "defaulting to standard logger.",
        )
        return StackdriverLogMetricsTimer(*args, **kwargs)


class StackdriverLogMetricsCounter(logger.LoggerCounter):
    """Stackdriver log-based counter metric.

    .. note::

        Stackdriver counts log lines so initializing a counter value is
        not supported .

    Args:
        name (str): name of counter
        job_name (str): name of Dataflow job
        project (str): name of GCP project associated with Dataflow job
        transform (str): Name of transform associated with counter, if any.
        tags (dict): Tags to associate with counter. Note:
            ``{"metric_type": "counter"}`` will always be an included tag.
    """

    # NOTE: The in-memory value (as kept track in the metric Dispatchers) may
    # be greater than 1, but it doesn't mean anything because:
    # 1. log-based counters in SD count the number log lines, and does not
    #    extract a value from the log line, and
    # 2. state is not maintained within a pipeline, so the in-memory
    #    value will eventually be wiped and re-initialized.
    # Therefore, we're hard-coding counters to the value of 1
    DEFAULT_LOG_FORMAT = (
        "[{name}] value: 1 transform: '{transform}' tags: {tags}"
    )
    KLIO_TRANSFORM_LABEL_KEY = "klio_transform"

    def __init__(self, name, job_name, project, transform=None, tags=None):
        # Since stackdriver literally counts loglines, initializing a
        # counter value is not supported; defaulting to 0
        super(StackdriverLogMetricsCounter, self).__init__(
            name, value=0, transform=transform, tags=tags
        )
        self.job_name = job_name
        self.project = project
        self.parent = "projects/{}".format(project)
        self.desc = "Klio counter '{}' ".format(self.name)
        self.body = self._get_body()

    def _get_filter(self):
        _filter = (
            'resource.type="dataflow_step" '
            'logName="projects/{project}/logs/'
            'dataflow.googleapis.com%2Fworker" '
            # since stackdriver counters do not support regexes, we're
            # including the [] around the name to avoid name collisions like
            # my-counter & my-counter2
            'jsonPayload.message:"[{name}]"'.format(
                project=self.project, name=self.name
            )
        )
        return _filter

    def _get_transform_label_extractor(self):
        # Grab the transform name from the log line, i.e.:
        # "[my-counter] value: 1 transform: 'HelloKlio' tags: {'metric_...";
        # Needs to be able to follow valid Python chars for class names:
        # https://stackoverflow.com/a/10120327/1579977
        label_extractor = "\"transform: '([a-zA-Z_][a-zA-Z0-9_]*)' tags:\""

        # docs: https://cloud.google.com/logging/docs/
        #              logs-based-metrics/labels#create-label
        label_regex = "REGEXP_EXTRACT(jsonPayload.message, {})".format(
            label_extractor
        )

        return {self.KLIO_TRANSFORM_LABEL_KEY: label_regex}

    def _get_body(self):
        labels = [
            {
                "key": self.KLIO_TRANSFORM_LABEL_KEY,
                "valueType": "STRING",
                "description": "Name of Klio-based transform",
            }
        ]
        body = {
            "name": self.name,
            "description": self.desc,
            "filter": self._get_filter(),
            "metricDescriptor": {
                "metricKind": "DELTA",
                "valueType": "INT64",
                "unit": "1",
                "labels": labels,
            },
            "labelExtractors": self._get_transform_label_extractor(),
        }
        return body

    def _init_metric(self, stackdriver_client):
        req = (
            stackdriver_client.projects()
            .metrics()
            .create(parent=self.parent, body=self.body)
        )
        try:
            req.execute()

        except gapi_errors.HttpError as e:
            # HTTP Conflict - metric already exists
            if e.resp.status == 409:
                logging.getLogger("klio.metrics").debug(
                    "Metric {} already exists".format(self.name)
                )
                return

            logging.getLogger("klio.metrics").error(
                "Error creating metric '{}': {}".format(self.name, e),
                exc_info=True,
            )

        except Exception as e:
            logging.getLogger("klio.metrics").error(
                "Error creating metric '{}': {}".format(self.name, e),
                exc_info=True,
            )

    # including method for documentation purposes


class StackdriverLogMetricsGauge(logger.LoggerGauge):
    """Pass-thru object for naming only. Stackdriver log-based metrics
    does not support gauges.
    """


class StackdriverLogMetricsTimer(logger.LoggerTimer):
    """Pass-thru object for naming only. Stackdriver log-based metrics
    does not support timers.
    """



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from apache_beam import pvalue

from klio_core.proto import klio_pb2

from klio.message import exceptions


def _handle_msg_compat(parsed_message):
    if parsed_message.version is klio_pb2.Version.V1:
        if parsed_message.data.entity_id and not parsed_message.data.element:
            # make v1 messages compatible with v2
            parsed_message.data.element = bytes(
                parsed_message.data.entity_id, "utf-8"
            )
        return parsed_message

    if parsed_message.version is klio_pb2.Version.V2:
        # is it safe to assume if a message is already labeled as v2, it should
        # have an element or payload? i.e. not just entity_id?
        return parsed_message

    if parsed_message.data.entity_id and not parsed_message.data.element:
        # assume v1 message
        parsed_message.version = klio_pb2.Version.V1
        # make v1 messages compatible with v2
        parsed_message.data.element = bytes(
            parsed_message.data.entity_id, "utf-8"
        )

    elif not parsed_message.data.entity_id and not parsed_message.data.element:
        # assume v1 message
        parsed_message.version = klio_pb2.Version.V1

    elif parsed_message.data.element and not parsed_message.data.entity_id:
        # assume v2 message
        parsed_message.version = klio_pb2.Version.V2

    return parsed_message


# [batch dev] attemping to make this a little generic so it can (eventually)
# be used with transforms other than DoFns
def to_klio_message(incoming_message, kconfig=None, logger=None):
    """Serialize ``bytes`` to a :ref:`KlioMessage <klio-message>`.

    .. tip::

        Set ``job_config.allow_non_klio_messages`` to ``True`` in
        ``klio-job.yaml`` in order to process non-``KlioMessages`` as
        regular ``bytes``. This function will create a new ``KlioMessage``
        and set the incoming ``bytes`` to ``KlioMessage.data.element``.

    Args:
        incoming_message (bytes): Incoming bytes to parse into a \
            ``KlioMessage``.
        kconfig (klio_core.config.KlioConfig): the current job's
            configuration.
        logger (logging.Logger): the logger associated with the Klio
            job.
    Returns:
        klio_core.proto.klio_pb2.KlioMessage: a ``KlioMessage``.
    Raises:
        klio_core.proto.klio_pb2._message.DecodeError: incoming message
            can not be parsed into a ``KlioMessage`` and
            ``job_config.allow_non_klio_messages`` in ``klio-job.yaml``
            is set to ``False``.
    """
    # TODO: when making a generic de/ser func, be sure to assert
    # kconfig and logger exists
    parsed_message = klio_pb2.KlioMessage()

    try:
        parsed_message.ParseFromString(incoming_message)

    except klio_pb2._message.DecodeError as e:
        if kconfig.job_config.allow_non_klio_messages:
            # We are assuming that we have been given "raw" data that is not in
            # the form of a serialized KlioMessage.
            parsed_message.data.element = incoming_message
            # default to set recipients to anyone - can't know who the
            # appropriate recipient is when it's not a real klio msg
            parsed_message.metadata.intended_recipients.anyone.SetInParent()
            parsed_message.version = klio_pb2.Version.V2
        else:
            logger.error(
                "Can not parse incoming message. To support non-Klio "
                "messages, add `job_config.allow_non_klio_messages = true` "
                "in the job's `klio-job.yaml` file."
            )
            raise e

    parsed_message = _handle_msg_compat(parsed_message)
    return parsed_message


def _handle_v2_payload(klio_message, payload):
    if payload:
        # if the user just returned exactly what they received in the
        # process method; let's avoid recursive payloads
        if payload == klio_message.data:
            payload = b""

    if not payload:
        # be sure to clear out old payload if there's no new payload
        payload = b""

    else:
        if not isinstance(payload, bytes):
            try:
                payload = bytes(payload, "utf-8")
            except TypeError:
                msg = (
                    "Returned payload could not be coerced to `bytes`.\n"
                    "Erroring payload: {}\nErroring KlioMessage: {}".format(
                        payload, klio_message
                    )
                )
                raise exceptions.KlioMessagePayloadException(msg)
    return payload


def from_klio_message(klio_message, payload=None):
    """Deserialize a given :ref:`KlioMessage <klio-message>` to ``bytes``.

    Args:
        klio_message (klio_core.proto.klio_pb2.KlioMessage): the
            ``KlioMessage`` in which to deserialize into ``bytes``
        payload (bytes or str): Optional ``bytes`` or ``str`` to update
            the value of ``KlioMessage.data.payload`` with before
            deserializing into bytes. Default: ``None``.
    Returns:
        bytes: a ``KlioMessage`` as ``bytes``.
    Raises:
        exceptions.KlioMessagePayloadException: the provided payload
            value cannot be coerced into ``bytes``.
    """
    tagged, tag = False, None
    if isinstance(payload, pvalue.TaggedOutput):
        tagged = True
        tag = payload.tag
        payload = payload.value

    # only update payload if it's a v2 message.
    if klio_message.version == klio_pb2.Version.V2:
        payload = _handle_v2_payload(klio_message, payload)
        # [batch dev] TODO: figure out how/where to clear out this payload
        # when publishing to pubsub (and potentially other output transforms)
        klio_message.data.payload = payload

    if tagged:
        return pvalue.TaggedOutput(tag, klio_message.SerializeToString())

    return klio_message.SerializeToString()



# Copyright 2019-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Base classes from which a metrics consumer (i.e. Stackdriver, ffwd, etc.)
will need to implement.

New consumers are required to implement the :class:`AbstractRelayClient`, and
three metrics objects based off of :class:`BaseMetric`: a counter, a gauge, and
a timer.
"""
import abc

import six


class _DummyAttribute(object):
    # for the ability to do `FOO_ATTR = abstract_attr()` as well as
    # decorate a property method
    pass


def abstract_attr(obj=None):
    """Set an attribute or a property as abstract.

    Supports class-level attributes as well as methods defined as a
    ``@property``.

    Usage:

    .. code-block:: python

        class Foo(object):
            my_foo_attribute = abstract_attr()

            @property
            @abstract_attr
            def my_foo_property(self):
                pass

    Args:
        obj (callable): Python object to "decorate", i.e. a class method. If
            none is provided, a dummy object is created in order to attach
            the ``__isabstractattr__`` attribute (similar to
            ``__isabstractmethod__`` from ``abc.abstractmethod``).

    Returns object with ``__isabstractattr__`` attribute set to ``True``.
    """
    if not obj:
        obj = _DummyAttribute()
    obj.__isabstractattr__ = True
    return obj


def _has_abstract_attributes_implemented(cls, name, bases):
    """Verify a given class has its abstract attributes implemented."""
    for base in bases:
        abstract_attrs = getattr(base, "_klio_metrics_abstract_attributes", [])
        class_attrs = getattr(cls, "_klio_metrics_all_attributes", [])
        for attr in abstract_attrs:
            if attr not in class_attrs:
                err_str = (
                    "Error instantiating class '{0}'. Implementation of "
                    "abstract attribute '{1}' from base class '{2}' is "
                    "required.".format(name, attr, base.__name__)
                )
                raise NotImplementedError(err_str)


def _get_all_attributes(clsdict):
    return [name for name, val in six.iteritems(clsdict) if not callable(val)]


def _get_abstract_attributes(clsdict):
    return [
        name
        for name, val in six.iteritems(clsdict)
        if not callable(val) and getattr(val, "__isabstractattr__", False)
    ]


class _ABCBaseMeta(abc.ABCMeta):
    """Enforce behavior upon implementations of ABC classes."""

    def __init__(cls, name, bases, clsdict):
        _has_abstract_attributes_implemented(cls, name, bases)

    def __new__(metaclass, name, bases, clsdict):
        clsdict[
            "_klio_metrics_abstract_attributes"
        ] = _get_abstract_attributes(clsdict)
        clsdict["_klio_metrics_all_attributes"] = _get_all_attributes(clsdict)
        cls = super(_ABCBaseMeta, metaclass).__new__(
            metaclass, name, bases, clsdict
        )
        return cls


class AbstractRelayClient(six.with_metaclass(_ABCBaseMeta)):
    """Abstract base class for all metric consumer relay clients.

    Each new consumer (i.e. Stackdriver, ffwd, logging-based metrics)
    will need to implement this relay class.

    Attributes:
        RELAY_CLIENT_NAME (str): must match the key in ``klio-job.yaml``
            under ``job_config.metrics``.

    """

    RELAY_CLIENT_NAME = abstract_attr()

    def __init__(self, klio_config):
        self.klio_config = klio_config

    @abc.abstractmethod
    def unmarshal(self, metric):
        """Returns a dictionary-representation of the ``metric`` object"""
        pass

    @abc.abstractmethod
    def emit(self, metric):
        """Emit the given metric object to the particular consumer.

        ``emit`` will be run in a threadpool separate from the transform,
        and any errors raised from the method will be logged then ignored.
        """
        pass

    @abc.abstractmethod
    def counter(self, name, value=0, transform=None, **kwargs):
        """Return a newly instantiated counter-type metric specific for
        the particular consumer.

        Callers to the ``counter`` method will store new counter objects
        returned in memory for simple caching.
        """
        pass

    @abc.abstractmethod
    def gauge(self, name, value=0, transform=None, **kwargs):
        """Return a newly instantiated gauge-type metric specific for
        the particular consumer.

        Callers to the ``gauge`` method will store new gauge objects
        returned in memory for simple caching.
        """
        pass

    @abc.abstractmethod
    def timer(self, name, transform=None, **kwargs):
        """Return a newly instantiated timer-type metric specific for
        the particular consumer.

        Callers to the ``timer`` method will store new timer objects
        returned in memory for simple caching.
        """
        pass


class BaseMetric(object):
    """Base class for all metric types.

    A consumer must implement a counter metric, a gauge metric, and a
    timer metric.
    """

    def __init__(self, name, value=0, transform=None, **kwargs):
        self.name = name
        self.value = value
        self.transform = transform

    def update(self, value):
        self.value = value



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import collections
import contextlib
import functools
import inspect
import threading
import types

import apache_beam as beam
from apache_beam import pvalue

from klio import utils as kutils
from klio.message import serializer
from klio.transforms import _retry as kretry
from klio.transforms import _timeout as ktimeout
from klio.transforms import _utils as txf_utils
from klio.transforms import core


_ERROR_MSG_KMSG_FROM_BYTES = (
    "Dropping KlioMessage - exception occurred when serializing '{}' "
    "from bytes to a KlioMessage.\n"
    "Error: {}"
)
_ERROR_MSG_KMSG_TO_BYTES = (
    "Dropping KlioMessage - exception occurred when deserializing '{}' "
    "from a KlioMessage to bytes.\n"
    "Error: {}"
)
_CACHED_INSTANCES = collections.defaultdict(set)


# TODO: This may be nice to make generic and move into the public helpers
# module since users may want to use something like this.
def __get_or_create_instance(function):
    """Creates a context manager so we can avoid creating instances of
    the same object (in our case, KlioContext).

    This uses a lock so we can avoid threading issues in Dataflow.
    """
    lock = threading.Lock()

    @contextlib.contextmanager
    def func(*args, **kwargs):
        key = "KlioContext_" + str(id(function))
        try:
            instance = _CACHED_INSTANCES[key].pop()
        except KeyError:
            with lock:
                instance = function()
        try:
            yield instance
        finally:
            _CACHED_INSTANCES[key].add(instance)

    return func


@__get_or_create_instance
def _klio_context():
    return core.KlioContext()


def __is_method(obj):
    # we have to do this instead of inspect.ismethod since
    # it's not a "bounded" method, ie Foo.process (unbounded)
    # vs Foo().process (bounded)
    args = inspect.getfullargspec(obj).args
    if args:
        return "self" == args[0]
    return False


def __is_dofn_process_method(self, meth):
    if meth.__name__ == "process":
        if issubclass(self.__class__, beam.DoFn):
            return True
    return False


def __get_user_error_message(err, func_path, kmsg):
    tb, exc_info = "", True

    if hasattr(err, "_klio_traceback"):
        tb = "\n" + err._klio_traceback
        exc_info = False

    if isinstance(err, ktimeout.KlioTimeoutError):
        # no need to include traceback of a timeout error
        exc_info = False

    msg = (
        "Dropping KlioMessage - exception occurred when calling '%s' with "
        "'%s'.\nError: %s%s" % (func_path, kmsg, err, tb)
    )
    return msg, exc_info


def __get_thread_limiter(max_thread_count, thread_limiter, func_name=None):
    if max_thread_count is not None and thread_limiter is not None:
        # raise a runtime error so it actually crashes klio/beam rather than
        # just continue processing elements
        raise RuntimeError(
            "`max_thread_count` and `thread_limiter` are mutually exclusive "
            "arguments."
        )

    if thread_limiter is not None:
        if not isinstance(thread_limiter, kutils.ThreadLimiter):
            # raise a runtime error so it actually crashes klio/beam rather
            # than just continue processing elements
            raise RuntimeError(
                "'thread_limiter' must be an instance of `klio.utils."
                "ThreadLimiter`."
            )

    if max_thread_count is not None:
        is_int_enum = isinstance(max_thread_count, (int, kutils.ThreadLimit))
        is_func = callable(max_thread_count)
        if not any([is_int_enum, is_func]):
            # raise a runtime error so it actually crashes klio/beam rather
            # than just continue processing elements
            raise RuntimeError(
                "Invalid type for handle_klio's argument 'max_thread_count'. "
                "Expected an `int`, a callable returning an `int`, or "
                "`klio.utils.ThreadLimit`, got `%s`."
                % type(max_thread_count).__name__
            )

        if isinstance(max_thread_count, int) and max_thread_count <= 0:
            # raise a runtime error so it actually crashes klio/beam rather
            # than just continue processing elements
            raise RuntimeError(
                "'max_thread_count' must be greater than 0. Set "
                "'max_thread_count' to `None` or `klio.utils.ThreadLimiter."
                "NONE` to turn off thread limitations."
            )
    if max_thread_count is None and thread_limiter is None:
        max_thread_count = kutils.ThreadLimit.DEFAULT

    if thread_limiter is None:
        thread_limiter = kutils.ThreadLimiter(
            max_thread_count=max_thread_count, name=func_name
        )

    return thread_limiter


# A separate function from __serialize_klio_message_generator so we can
# specifically `yield from` it (and exhaust transforms that have multiple
# yields)
def __from_klio_message_generator(self, kmsg, payload, orig_item):
    try:
        yield serializer.from_klio_message(kmsg, payload)

    except Exception as err:
        self._klio.logger.error(
            _ERROR_MSG_KMSG_TO_BYTES.format(kmsg, err), exc_info=True
        )
        # Since the yielded value in the `try` clause may not tagged, that
        # one will be used by default by whatever executed this function,
        # and anything that has a tagged output value (like this dropped one)
        # will just be ignored, which is fine for dropped values.
        # But if the caller function wanted to, they could access this via
        # pcoll.drop.
        # We won't try to serialize kmsg to bytes since something already
        # went wrong.
        yield pvalue.TaggedOutput("drop", orig_item)
        # explicitly return so that Beam doesn't call `next` and
        # executes the next `yield`
        return


# meant for DoFn.process generator methods; very similar to the
# '__serialize_klio_message' function, but needs to be its own function
# since it's a generator (can't mix the two - if a function has a `yield`
# statement, it automatically becomes a generator)
def __serialize_klio_message_generator(
    self, meth, incoming_item, *args, **kwargs
):
    try:
        kmsg = serializer.to_klio_message(
            incoming_item, self._klio.config, self._klio.logger
        )
    except Exception as err:
        self._klio.logger.error(
            _ERROR_MSG_KMSG_FROM_BYTES.format(incoming_item, err),
            exc_info=True,
        )
        # Since the yielded value in the `try` clause is not tagged, that
        # one will be used by default by whatever executed this function,
        # and anything that has a tagged output value (like this dropped one)
        # will just be ignored, which is fine for dropped values.
        # But if the caller function wanted to, they could access this via
        # pcoll.drop.
        yield pvalue.TaggedOutput("drop", incoming_item)
        # explicitly return so that Beam doesn't call `next` and
        # executes the next `yield`
        return

    try:
        payload = meth(self, kmsg.data, *args, **kwargs)

    except Exception as err:
        func_path = self.__class__.__name__ + "." + meth.__name__
        log_msg, exc_info = __get_user_error_message(err, func_path, kmsg)
        self._klio.logger.error(log_msg, exc_info=exc_info)
        # Since the yielded value in the `try` clause is not tagged, that
        # one will be used by default by whatever executed this function,
        # and anything that has a tagged output value (like this dropped one)
        # will just be ignored, which is fine for dropped values.
        # But if the caller function wanted to, they could access this via
        # pcoll.drop.
        # We won't try to serialize kmsg to bytes since something already
        # went wrong.
        yield pvalue.TaggedOutput("drop", incoming_item)
        # explicitly return so that Beam doesn't call `next` and
        # executes the next `yield`
        return

    else:
        if isinstance(payload, types.GeneratorType):
            try:
                for pl in payload:
                    yield from __from_klio_message_generator(
                        self, kmsg, pl, incoming_item
                    )
            # This exception block will the execute
            # if the pl item is an Exception
            except Exception as err:
                func_path = self.__class__.__name__ + "." + meth.__name__
                log_msg, exc_info = __get_user_error_message(
                    err, func_path, kmsg
                )
                self._klio.logger.error(log_msg, exc_info=exc_info)
                # This will catch an exception present in the generator
                # containing items yielded by a function/method
                # decorated by @handle_klio.
                # Following items in the generator will be ignored
                # since an exception has already been detected.
                # We won't try to serialize kmsg to bytes since
                # something already went wrong.
                yield pvalue.TaggedOutput("drop", incoming_item)
                # explicitly return so that Beam doesn't call `next` and
                # executes the next `yield`
                return
        else:
            yield from __from_klio_message_generator(
                self, kmsg, payload, incoming_item
            )


def __serialize_klio_message(ctx, func, incoming_item, *args, **kwargs):
    # manipulate `ctx` to handle both methods and functions depending on
    # what we're wrapping. Functions just have `ctx` object, but methods
    # have `self._klio` as its context, and we also need access to `self`
    # in order to call the method
    _self = ctx
    if not isinstance(ctx, core.KlioContext):
        ctx = _self._klio
    try:
        kmsg = serializer.to_klio_message(
            incoming_item, ctx.config, ctx.logger
        )
    except Exception as err:
        ctx.logger.error(
            _ERROR_MSG_KMSG_FROM_BYTES.format(incoming_item, err),
            exc_info=True,
        )
        # Since the returned value in the `try` clause is not tagged, that
        # one will be used by default by whatever executed this function,
        # and anything that has a tagged output value (like this dropped one)
        # will just be ignored, which is fine for dropped values.
        # But if the caller function wanted to, they could access this via
        # pcoll.drop.
        return pvalue.TaggedOutput("drop", incoming_item)

    try:
        ret = func(_self, kmsg.data, *args, **kwargs)
        if isinstance(ret, types.GeneratorType):
            raise TypeError(
                "can't pickle generator object: '{}'".format(func.__name__)
            )
    except TypeError:
        # If we get here, we threw a type error because we found a generator
        # and those can't be pickled. But there's no need to do any special
        # error handling - this will contain enough info for the user so
        # we just re-raise
        raise

    except Exception as err:
        log_msg, exc_info = __get_user_error_message(err, func.__name__, kmsg)
        ctx.logger.error(log_msg, exc_info=exc_info)
        # Since the returned value in the `try` clause is not tagged, that
        # one will be used by default by whatever executed this function,
        # and anything that has a tagged output value (like this dropped one)
        # will just be ignored, which is fine for dropped values.
        # But if the caller function wanted to, they could access this via
        # pcoll.drop.
        # We won't try to serialize kmsg to bytes since something already
        # went wrong.
        return pvalue.TaggedOutput("drop", incoming_item)

    try:
        return serializer.from_klio_message(kmsg, ret)

    except Exception as err:
        ctx.logger.error(
            _ERROR_MSG_KMSG_TO_BYTES.format(kmsg, err), exc_info=True
        )
        # Since the returned value in the `try` clause is not tagged, that
        # one will be used by default by whatever executed this function,
        # and anything that has a tagged output value (like this dropped one)
        # will just be ignored, which is fine for dropped values.
        # But if the caller function wanted to, they could access this via
        # pcoll.drop.
        # We won't try to serialize kmsg to bytes since something already
        # went wrong.
        return pvalue.TaggedOutput("drop", incoming_item)


def _serialize_klio_message(func_or_meth):
    @functools.wraps(func_or_meth)
    def method_wrapper(self, incoming_item, *args, **kwargs):
        wrapper = __serialize_klio_message
        if __is_dofn_process_method(self, func_wrapper):
            wrapper = __serialize_klio_message_generator

        return wrapper(self, func_or_meth, incoming_item, *args, **kwargs)

    @functools.wraps(func_or_meth)
    def func_wrapper(klio_ns, incoming_item, *args, **kwargs):
        return __serialize_klio_message(
            func_or_meth, klio_ns, incoming_item, *args, **kwargs
        )

    if __is_method(func_or_meth):
        return method_wrapper
    return func_wrapper


def _set_klio_context(method):
    @functools.wraps(method)
    def method_wrapper(self, *args, **kwargs):
        if not getattr(self, "_klio", None):
            with _klio_context() as ctx:
                setattr(self, "_klio", ctx)
        return method(self, *args, **kwargs)

    return method_wrapper


def _inject_klio_context(func_or_meth):
    @functools.wraps(func_or_meth)
    def method_wrapper(self, *args, **kwargs):
        with _klio_context() as ctx:
            return func_or_meth(self, ctx, *args, **kwargs)

    @functools.wraps(func_or_meth)
    def func_wrapper(*args, **kwargs):
        with _klio_context() as ctx:
            return func_or_meth(ctx, *args, **kwargs)

    if __is_method(func_or_meth):
        return method_wrapper
    return func_wrapper


def _handle_klio(*args, max_thread_count=None, thread_limiter=None, **kwargs):
    def inner(func_or_meth):
        func_name = getattr(
            func_or_meth, "__qualname__", func_or_meth.__name__
        )
        thd_limiter = __get_thread_limiter(
            max_thread_count, thread_limiter, func_name
        )

        @functools.wraps(func_or_meth)
        def method_wrapper(self, *args, **kwargs):
            with thd_limiter:
                with _klio_context() as ctx:
                    setattr(self, "_klio", ctx)

                # SO. HACKY. We check to see if this method is named "expand"
                # to designate  if the class is a Composite-type transform
                # (rather than a DoFn with a "process" method).
                # A Composite transform handles a pcoll / pipeline,
                # not the individual elements, and therefore doesn't need
                # to be given a KlioMessage. It should only need the KlioContext
                # attached.
                if func_or_meth.__name__ == "expand":
                    return func_or_meth(self, *args, **kwargs)

                wrapper = __serialize_klio_message
                # Only the process method of a DoFn is a generator - otherwise
                # beam can't pickle a generator
                if __is_dofn_process_method(self, func_or_meth):
                    wrapper = __serialize_klio_message_generator

                incoming_item = args[0]
                args = args[1:]
                return wrapper(
                    self, func_or_meth, incoming_item, *args, **kwargs
                )

        @functools.wraps(func_or_meth)
        def func_wrapper(incoming_item, *args, **kwargs):
            with thd_limiter:
                with _klio_context() as ctx:
                    return __serialize_klio_message(
                        ctx, func_or_meth, incoming_item, *args, **kwargs
                    )

        if __is_method(func_or_meth):
            return method_wrapper
        return func_wrapper

    # allows @handle_klio to be used without parens (i.e. no need to do
    # `@handle_klio()`) when there are no args/kwargs provided
    if args and callable(args[0]):
        return inner(args[0])
    return inner


def _timeout(seconds=None, exception=None, exception_message=None):
    try:
        seconds = float(seconds)
    except ValueError:
        # raise a runtime error so it actually crashes klio/beam rather than
        # just continue processing elements
        raise RuntimeError(
            "Invalid type for timeout 'seconds'. Expected a `float` or an "
            "`int`, got `%s`." % type(seconds).__name__
        )

    if not seconds > 0:
        # raise a runtime error so it actually crashes klio/beam rather than
        # just continue processing elements
        raise RuntimeError(
            "Invalid value '%d' for timeout. Must be a positive number."
            % seconds
        )

    def inner(func_or_meth):
        timeout_wrapper = ktimeout.KlioTimeoutWrapper(
            function=func_or_meth,
            seconds=seconds,
            timeout_exception=exception,
            exception_message=exception_message,
        )

        # Unfortunately these two wrappers can't be abstracted into
        # one wrapper - the `self` arg apparently can not be abstracted
        @functools.wraps(func_or_meth)
        def method_wrapper(self, kmsg, *args, **kwargs):
            args = (self, kmsg) + args
            return timeout_wrapper(*args, **kwargs)

        @functools.wraps(func_or_meth)
        def func_wrapper(ctx, kmsg, *args, **kwargs):
            args = (ctx, kmsg) + args
            return timeout_wrapper(*args, **kwargs)

        if __is_method(func_or_meth):
            return method_wrapper
        return func_wrapper

    return inner


def _retry(
    *args,
    tries=None,
    delay=None,
    exception=None,
    raise_exception=None,
    exception_message=None,
    **kwargs
):
    if tries is None:
        tries = -1  # infinite

    if not isinstance(tries, int):
        # raise a runtime error so it actually crashes klio/beam rather than
        # just continue processing elements
        raise RuntimeError(
            "Invalid type for retry 'tries'. Expected an int, got "
            "`{}`.".format(type(tries).__name__)
        )

    if tries < -1:
        # raise a runtime error so it actually crashes klio/beam rather than
        # just continue processing elements
        raise RuntimeError(
            "Invalid value '%d' for retry 'tries'. Must be a positive integer "
            "or '-1' for infinite tries." % tries
        )

    if delay is None:
        delay = 0

    if not isinstance(delay, (int, float)):
        raise RuntimeError(
            "Invalid type for retry 'delay'. Expected a `float` or an "
            "`int`, got `%s`." % type(delay).__name__
        )

    if delay < 0:
        # raise a runtime error so it actually crashes klio/beam rather than
        # just continue processing elements
        raise RuntimeError(
            "Invalid value '%d' for retry 'delay'. Must be a positive number. "
            % delay
        )

    def inner(func_or_meth):
        retry_wrapper = kretry.KlioRetryWrapper(
            function=func_or_meth,
            tries=tries,
            delay=delay,
            exception=exception,
            raise_exception=raise_exception,
            exception_message=exception_message,
        )

        # Unfortunately these two wrappers can't be abstracted into
        # one wrapper - the `self` arg apparently can not be abstracted
        @functools.wraps(func_or_meth)
        def method_wrapper(self, kmsg, *args, **kwargs):
            args = (self, kmsg) + args
            return retry_wrapper(*args, **kwargs)

        @functools.wraps(func_or_meth)
        def func_wrapper(ctx, kmsg, *args, **kwargs):
            args = (ctx, kmsg) + args
            return retry_wrapper(*args, **kwargs)

        if __is_method(func_or_meth):
            return method_wrapper
        return func_wrapper

    return inner


# Allow internals to call semiprivate funcs without triggering
# user-facing warnings
@txf_utils.experimental()
def serialize_klio_message(*args, **kwargs):
    """Serialize/deserialize incoming PCollections as a KlioMessage.

    This decorator needs access to a :class:`KlioContext
    <klio.transforms.core.KlioContext>` object via ``@inject_klio_context``
    or ``@set_klio_context`` if not available on the object (i.e. ``self`` of
    a DoFn instance), or use ``@handle_klio`` which will handle ``KlioContext``
    and KlioMessage de/serialization.
    """
    return _serialize_klio_message(*args, **kwargs)


@txf_utils.experimental()
def set_klio_context(*args, **kwargs):
    """Set :class:`KlioContext <klio.transforms.core.KlioContext>` to the
    class instance.

    Use ``@handle_klio`` instead if KlioMessage de/serialization is
    also needed.

    Use ``@inject_klio_context`` if using on a function rather than a
    class method.

    .. code-block:: python

        class MyComposite(beam.PTransform):
            @set_klio_context
            def expand(self, element):
                self._klio.logger.info(f"Received {element}")
    """
    return _set_klio_context(*args, **kwargs)


@txf_utils.experimental()
def inject_klio_context(*args, **kwargs):
    """Provide :class:`KlioContext <klio.transforms.core.KlioContext>` as the
    first argument to a decorated method/func.

    Use ``@handle_klio`` instead if KlioMessage de/serialization is
    also needed.

    If not needing KlioMessage de/serialization, consider
    ``@set_klio_context`` if using on a class method (rather than a
    function) to set the ``_klio`` attribute on ``self``.

    .. code-block:: python

        @inject_klio_context
        def my_map_func(ctx, element):
            ctx.logger.info(f"Received {element}")

        class MyDoFn(beam.DoFn):
            @inject_klio_context
            def process(self, ctx, element):
                ctx.logger.info(f"Received {element}")
    """
    return _inject_klio_context(*args, **kwargs)


# TODO: Update docstrings w/ new kwargs & examples
@txf_utils.experimental()
def handle_klio(*args, max_thread_count=None, thread_limiter=None, **kwargs):
    """Serialize & deserialize incoming PCollections as a KlioMessage.

    Behind the scenes, this generates :class:`KlioContext
    <klio.transforms.core.KlioContext>`, handles de/serialize the
    incoming PCollection as a Klio Message, as well as manage thread
    concurrency.

    .. admonition:: Default Thread Concurrency Management
        :class: caution

        The ``@handle_klio`` decorator will default to limiting the amount
        of active threads a decorated transform can use. The default
        maximum number of active threads is the number of CPUs on the
        worker machine.

        See examples below on how to adjust this behavior.

        Learn more about Klio's thread concurrency management
        :ref:`in the User Guide <klio-concurrency-mgmt>`.

    If decorating a class method, the ``KlioContext`` will be attached
    to the ``self`` argument of the class instance.

    If decorating a function, ``KlioContext`` will be provided as the first
    argument.

    .. code-block:: python

        @handle_klio
        def my_map_func(ctx, item):
            ctx.logger.info(f"Received {item.element} with {item.payload}")

        class MyDoFn(beam.DoFn):
            @handle_klio
            def process(self, item):
                self._klio.logger.info(
                    f"Received {item.element} with {item.payload}"
                )

        class MyComposite(beam.PTransform):
            @handle_klio
            def expand(self, pcoll):
                kms_config = self._klio.config.job_config.kms_config
                return pcoll | MyKMSTransform(**kms_config)

    To adjust the maximum threads a decorated transform uses:

    .. code-block:: python

        from klio import utils as klio_utils

        # Set the limit to 4 threads
        @handle_klio(max_thread_count=4):
        def my_map_func(ctx, item):
            ...

        # Set the limit to 2x CPU count
        import multiprocessing
        @handle_klio(max_thread_count=lambda: 2 * multiprocessing.cpu_count()):
        def my_map_func(ctx, item):
            ...

        # Turn off any thread limits
        @handle_klio(max_thread_count=klio_utils.ThreadLimit.NONE):
        def my_map_func(ctx, item):
            ...

        # Explicitly set the limit to Klio's default
        @handle_klio(max_thread_count=klio_utils.ThreadLimit.DEFAULT):
        def my_map_func(ctx, item):
            ...

        # Share thread limits between multiple transforms
        global_thread_limiter = klio_utils.ThreadLimiter(max_thread_count=4)

        @handle_klio(thread_limiter=global_thread_limiter)
        def first_map_func(ctx, item):
            ...

        @handle_klio(thread_limiter=global_thread_limiter)
        def second_map_func(ctx, item):
            ...

    Args:
        max_thread_count (int, callable, klio.utils.ThreadLimit): number of
            threads to make available to the decorated function, or a
            :func:`callable` that returns an ``int``. Set to
            :attr:`klio.utils.ThreadLimit.NONE` for no thread limits.
            Defaults to :attr:`klio.utils.ThreadLimit.DEFAULT` (worker CPU
            count) if ``thread_limiter`` is not provided.
            **Mutually exclusive** with ``thread_limiter`` argument.

        thread_limiter (klio.utils.ThreadLimiter): the ``ThreadLimiter``
            instance that the decorator should use instead of creating its own.
            Defaults to ``None``. **Mutually exclusive** with
            ``max_thread_count``.
    """
    return _handle_klio(
        *args,
        max_thread_count=max_thread_count,
        thread_limiter=thread_limiter,
        **kwargs,
    )


@txf_utils.experimental()
def timeout(seconds, *args, exception=None, exception_message=None, **kwargs):
    """Run the decorated method/function with a timeout in a separate process.

    If being used with :func:`@retry <retry>`, order is important depending
    on the desired effect. If ``@timeout`` is applied to a function before
    ``@retry``, then retries will apply first, meaning the configured timeout
    will cancel the function even if the retries have not yet been exhausted.
    In this case, be careful with the ``delay`` argument for the ``@retry``
    decorator: the set timeout is inclusive of a retry's delay. Conversely,
    if ``@retry`` is applied to a function before ``@timeout``, retries will
    continue until exhausted even if a function has timed out.

    If being used with another Klio decorator like :func:`@handle_klio
    <handle_klio>`, then the ``@timeout`` decorator should be applied to a
    method/function **after** another Klio decorator.

    .. code-block:: python

        @handle_klio
        @timeout(seconds=5)
        def my_map_func(ctx, item):
            ctx.logger.info(f"Received {item.element} with {item.payload}")

        class MyDoFn(beam.DoFn):
            @handle_klio
            @timeout(seconds=5, exception=MyTimeoutException)
            def process(self, item):
                self._klio.logger.info(
                    f"Received {item.element} with {item.payload}"
                )

        @timeout(
            seconds=5,
            exception=MyTimeoutException,
            exception_message="I got a timeout!"
        )
        def my_nonklio_map_func(item):
            print(f"Received {item}!")

    Args:
        seconds (float): The timeout period in seconds. Must be greater than 0.
        exception (Exception): The Exception that will be raised if a
            timeout occurs. Default: ``KlioTimeoutError``.
        exception_message (str): Custom exception message. Default:
            ``Function '{function}' timed out after {seconds} seconds.``
    """
    return _timeout(
        *args,
        seconds=seconds,
        exception=exception,
        exception_message=exception_message,
        **kwargs,
    )


@txf_utils.experimental()
def retry(
    *args,
    tries=None,
    delay=None,
    exception=None,
    raise_exception=None,
    exception_message=None,
    **kwargs
):
    """Retry a decorated method/function on failure.

    If being used with :func:`@timeout <timeout>`, order is important
    depending on the desired effect. If ``@timeout`` is applied to a function
    before ``@retry``, then retries will apply first, meaning the configured
    timeout will cancel the function even if the retries have not yet been
    exhausted. In this case, be careful with the ``delay`` argument for the
    ``@retry`` decorator: the set timeout is inclusive of a retry's delay.
    Conversely, if ``@retry`` is applied to a function before ``@timeout``,
    retries will continue until exhausted even if a function has timed out.

    If being used with a Klio decorator like :func:`@handle_klio
    <handle_klio>`, then the ``@retry`` decorator should be applied to a
    method/function **after** another Klio decorator.

    .. code-block:: python

        @handle_klio
        @retry()
        def my_map_func(ctx, item):
            ctx.logger.info(f"Received {item.element} with {item.payload}")

        class MyDoFn(beam.DoFn):
            @handle_klio
            @retry(tries=3, exception=MyExceptionToCatch)
            def process(self, item):
                self._klio.logger.info(
                    f"Received {item.element} with {item.payload}"
                )

        @retry(
            tries=3,
            exception=MyExceptionToCatch,
            raise_exception=MyExceptionToRaise,
            exception_message="All retries have been exhausted!"
        )
        def my_nonklio_map_func(item):
            print(f"Received {item}!")

    Args:
        tries (int): Maximum number of attempts. Default: -1 (infinite)
        delay (int or float): Delay between attempts in seconds. Default: 0
        exception (Exception or tuple(Exception)): An Exception or tuple of
            exceptions to catch and retry on. Defaults to ``Exception``.
        raise_exception (Exception): The ``Exception`` to raise once configured
            retries are exhausted. Defaults to ``KlioRetriesExhausted``.
        exception_message (str): Custom message for `raise_exception`.
            Default: ``Function '{}' has reached the maximum {} retries. Last
            exception: {}``

    """
    # There isn't a way to nicely support `@retry` because it might be used
    # with other decorators which adds a lot of complexity, so it must be
    # `@retry()`. But we can detect it and give a more friendly error.
    if len(args) == 1 and callable(args[0]):
        func_name = getattr(args[0], "__qualname__", args[0].__name__)

        raise RuntimeError(
            "The `retry` decorator needs to be called with parens on '{}', "
            "e.g. `@retry()`.".format(func_name)
        )

    return _retry(
        *args,
        tries=tries,
        delay=delay,
        exception=exception,
        raise_exception=raise_exception,
        exception_message=exception_message,
        **kwargs,
    )


# this is set by the exec profile command when it's time to profile stuff.
# When set, this should be a 1-arg function that takes another function as
# input and returns a wrapped version of it with profiling
ACTIVE_PROFILER = None


@txf_utils.experimental()
def profile(func_or_meth):
    """Decorator to mark a function/method for profiling.  This is used in
    conjunction with the ``klio job profile`` commands to selectively profile
    parts of your pipeline.  This decorator can be added to any function or
    method, but when using with other Klio decorators such as ``@handle_klio``
    it **must** be the last decorator applied.

    When running/testing a job normally and not profiling, this decorator has
    no effect.

    .. code-block:: python

        @handle_klio
        @profile
        def my_map_func(ctx, item):
            ctx.logger.info(f"Received {item.element} with {item.payload}")

        class MyDoFn(beam.DoFn):
            @handle_klio
            @profile
            def process(self, item):
                self._klio.logger.info(
                    f"Received {item.element} with {item.payload}"
                )

        @profile
        def my_nonklio_map_func(item):
            print(f"Received {item}!")
    """

    if ACTIVE_PROFILER is not None:
        decorated = ACTIVE_PROFILER(func_or_meth)

        @functools.wraps(func_or_meth)
        def inner_method(self, *args, **kwargs):
            return decorated(self, *args, **kwargs)

        @functools.wraps(func_or_meth)
        def inner_func(*args, **kwargs):
            return decorated(*args, **kwargs)

        # This is needed even though we treat them the same, other decorators
        # need to do the same check and won't work if we don't propagate the
        # `self` arg (which is not done by functools.wraps)
        if __is_method(func_or_meth):
            return inner_method
        else:
            return inner_func

    return func_or_meth



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import enum
import functools
import os

import apache_beam as beam

from apache_beam import pvalue
from apache_beam.io.gcp import gcsio

from klio.message import serializer
from klio.transforms import _utils
from klio.transforms import core


class DataExistState(enum.Enum):
    FOUND = "found"
    NOT_FOUND = "not_found"

    # to human-friendly strings for easy logging
    @classmethod
    def to_str(cls, attr):
        if attr == cls.NOT_FOUND:
            return "not found"
        return "found"


# TODO mabe use klio core config's defs but make them strings
class KlioIODirection(enum.Enum):
    INPUT = "input"
    OUTPUT = "output"


# TODO: maybe use common.MessageState which then should be mapped to
# strings instead of ints (or have a method that does the conversion)
class TaggedStates(enum.Enum):
    PROCESS = "process"
    PASS_THRU = "pass_thru"
    DROP = "drop"
    DEFAULT = "tag_not_set"


class KlioConfigRuntimeError(RuntimeError):
    """Config error that should crash the job at startup."""

    pass


class KlioUnsupportedConfigError(KlioConfigRuntimeError):
    """User-defined config is unsupported."""

    pass


# Only serializes to a KlioMessage; we deserialize within the process
# method itself since we also have to tag the output (too difficult to
# serialize output that's already tagged)
def _wrap_process(meth):
    @functools.wraps(meth)
    def wrapper(self, incoming_item, *args, **kwargs):
        try:
            kmsg = serializer.to_klio_message(
                incoming_item, self._klio.config, self._klio.logger
            )
            yield from meth(self, kmsg, *args, **kwargs)

        except Exception as err:
            self._klio.logger.error(
                "Dropping KlioMessage - exception occurred when serializing "
                "'%s' to a KlioMessage.\nError: %s" % (incoming_item, err),
                exc_info=True,
            )
            return

    return wrapper


def _job_in_jobs(current_job, job_list):
    # Use job name & project to ensure uniqueness
    curr_job_name = "{}-{}".format(
        current_job.gcp_project, current_job.job_name
    )
    downstream_job_names = [
        "{}-{}".format(j.gcp_project, j.job_name) for j in job_list
    ]

    return curr_job_name in downstream_job_names


class _KlioBaseDoFnMetaclass(type):
    """Enforce behavior upon subclasses of `_KlioBaseDataExistenceCheck`."""

    def __init__(cls, name, bases, clsdict):
        if not getattr(cls, "_klio", None):
            setattr(cls, "_klio", core.KlioContext())

        if os.getenv("KLIO_TEST_MODE", "").lower() in ("true", "1"):
            return

        # TODO: fixme: not every child class will inherit from
        # _KlioBaseDataExistenceCheck
        if _utils.is_original_process_func(
            clsdict, bases, base_class="_KlioBaseDataExistenceCheck"
        ):

            setattr(cls, "process", _wrap_process(clsdict["process"]))

            cls._klio._transform_name = name

    def __call__(self, *args, **kwargs):
        # automatically wrap DoFn in a beam.ParDo (with or without
        # `with_outputs` for tagged outputs) so folks can just do
        # `pcoll | KlioInputDataExistenceCheck()` rather than
        # `pcoll | beam.ParDo(KlioInputDataExistenceCheck()).with_outputs()`
        if self.WITH_OUTPUTS is True:
            return beam.ParDo(
                super(_KlioBaseDoFnMetaclass, self).__call__(*args, **kwargs)
            ).with_outputs()

        return beam.ParDo(
            super(_KlioBaseDoFnMetaclass, self).__call__(*args, **kwargs)
        )


class _KlioBaseDataExistenceCheck(beam.DoFn, metaclass=_KlioBaseDoFnMetaclass):
    """Base class for data existence checking."""

    DIRECTION_PFX = None  # i.e. KlioIODirection.INPUT
    WITH_OUTPUTS = True

    @property
    def _location(self):
        return self._data_config.location

    @property
    def _suffix(self):
        return self._data_config.file_suffix

    @property
    def _data_config(self):
        pass

    def exists(self, *args, **kwargs):
        pass

    def _get_absolute_path(self, element):
        return os.path.join(
            self._location, element.decode("utf-8") + self._suffix
        )


class _KlioInputDataMixin(object):
    """Mixin to add input-specific logic for a data existence check.

    Must be used with _KlioGcsCheckExistsBase
    """

    DIRECTION_PFX = KlioIODirection.INPUT

    @property
    def _data_config(self):
        # TODO: figure out how to support multiple inputs

        # If folks use the default existence checks that klio does, we
        # shouldn't get here. But we could if the user implements their
        # own and misconfigures their data inputs.
        if len(self._klio.config.job_config.data.inputs) > 1:
            # raise a runtime error so it actually crashes klio/beam rather than
            # just continue processing elements
            raise KlioUnsupportedConfigError(
                "Multiple inputs configured in "
                "`klio-job.yaml::job_config.data.inputs` are not supported "
                "for data existence checks."
            )
        # If folks use the default existence checks that klio does, we
        # shouldn't get here. But we could if the user implements their
        # own and misconfigures their data inputs.
        if len(self._klio.config.job_config.data.inputs) == 0:
            # raise a runtime error so it actually crashes klio/beam rather than
            # just continue processing elements
            raise KlioConfigRuntimeError(
                "Input data existence checks require input data to be "
                "configured in `klio-job.yaml::job_config.data.inputs`."
            )
        return self._klio.config.job_config.data.inputs[0]


class _KlioOutputDataMixin(object):
    """Mixin to add output-specific logic for a data existence check.

    Must be used with _KlioGcsCheckExistsBase
    """

    DIRECTION_PFX = KlioIODirection.OUTPUT

    @property
    def _data_config(self):
        # TODO: figure out how to support multiple outputs

        # If folks use the default existence checks that klio does, we
        # shouldn't get here. But we could if the user implements their
        # own and misconfigures their data outputs.
        if len(self._klio.config.job_config.data.outputs) > 1:
            # raise a runtime error so it actually crashes klio/beam rather than
            # just continue processing elements
            raise KlioUnsupportedConfigError(
                "Multiple outputs configured in "
                "`klio-job.yaml::job_config.data.outputs` are not supported "
                "for data existence checks."
            )

        # If folks use the default existence checks that klio does, we
        # shouldn't get here. But we could if the user implements their
        # own and misconfigures their data outputs.
        if len(self._klio.config.job_config.data.outputs) == 0:
            # raise a runtime error so it actually crashes klio/beam rather than
            # just continue processing elements
            raise KlioConfigRuntimeError(
                "Output data existence checks require output data to be "
                "configured in `klio-job.yaml::job_config.data.outputs`."
            )
        return self._klio.config.job_config.data.outputs[0]


class _KlioGcsDataExistsMixin(object):
    """Mixin for GCS-specific data existence check logic.

    Must be used with _KlioBaseDataExistenceCheck and either
    _KlioInputDataMixin or _KlioOutputDataMixin
    """

    def setup(self):
        self.client = gcsio.GcsIO()

    def exists(self, path):
        return self.client.exists(path)


class _KlioGcsCheckExistsBase(
    _KlioGcsDataExistsMixin, _KlioBaseDataExistenceCheck
):
    """Must be used with either _KlioInputDataMixin or _KlioOutputDataMixin"""

    def process(self, kmsg):
        item = kmsg.data.element
        item_path = self._get_absolute_path(item)
        item_exists = self.exists(item_path)

        state = DataExistState.FOUND
        if not item_exists:
            state = DataExistState.NOT_FOUND

        self._klio.logger.info(
            "%s %s at %s"
            % (
                self.DIRECTION_PFX.value.title(),
                DataExistState.to_str(state),
                item_path,
            )
        )

        # double tag for easier user interface, i.e. pcoll.found vs pcoll.true
        yield pvalue.TaggedOutput(state.value, kmsg.SerializeToString())



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module is heavily inspired by https://github.com/pnpnpn/timeout-decorator
and https://github.com/bitranox/wrapt_timeout_decorator.
It's been simplified for what we need, and adapted to work with generators.
"""

import multiprocessing
import sys
import time
import traceback
import types

import psutil


TRACEBACK_HEADER = "Traceback (most recent call last):\n"


class KlioTimeoutError(Exception):
    """Transform timed out while processing a KlioMessage."""


def _target(queue, function, *args, **kwargs):
    """Run user's function with args/kwargs and return output via a queue."""
    try:
        ret = function(*args, **kwargs)
        if isinstance(ret, types.GeneratorType):
            ret = next(ret)
        queue.put((True, ret))

    except Exception:
        # Note: here we're trying to get the traceback context that's relevant
        # to the user. Without it, the user would see the timeout wrapper
        # context, which won't help the user. So we grab the previous traceback
        # (accessed by tb_next) and attach it to the exception as a new
        # attribute. Then when we catch the exception in the caller code, we
        # can check for that attribute and log it.
        exc_infos = sys.exc_info()
        user_exc = exc_infos[1]
        user_exc_tb = "".join(traceback.format_tb(exc_infos[2].tb_next))
        user_exc_tb = TRACEBACK_HEADER + user_exc_tb
        setattr(user_exc, "_klio_traceback", user_exc_tb)
        queue.put((False, user_exc))


class KlioTimeoutWrapper(object):
    """Wrap a function to execute in a separate process with a timeout."""

    DEFAULT_EXC_MSG = "Function '{}' timed out after {} seconds."

    def __init__(
        self, function, seconds, timeout_exception=None, exception_message=None
    ):
        self._function = function
        self._func_name = getattr(function, "__qualname__", function.__name__)
        self._seconds = seconds
        self._timeout_exception = timeout_exception or KlioTimeoutError
        self._exception_message = exception_message

    def __call__(self, *args, **kwargs):
        self._queue = multiprocessing.Queue(maxsize=1)
        args = (self._queue, self._function) + args
        proc_name = "KlioTimeoutProcess-{}".format(self._func_name)
        self._process = multiprocessing.Process(
            target=_target, name=proc_name, args=args, kwargs=kwargs
        )
        # can't daemonize process in case users nest decorators
        self._process.daemon = False
        self._process.start()
        self._timeout = self._seconds + time.monotonic()

        while not self.ready:
            time.sleep(0.01)
        return self.value

    def _raise_exception(self):
        if self._exception_message is None:
            self._exception_message = self.DEFAULT_EXC_MSG.format(
                self._func_name, self._seconds
            )
        raise self._timeout_exception(self._exception_message)

    def _terminate_child_processes(self):
        # This is to make sure we clean up any spawned child processes,
        # particularly if the user uses `subprocess` in their timeout-
        # decorated code
        proc = psutil.Process(pid=self._process.pid)
        children = proc.children(recursive=True)
        while children:
            # top-down/handle parents first; if we terminate children first,
            # the parents could just respawn
            child = children.pop(0)
            try:
                child.terminate()
            except psutil.NoSuchProcess:
                pass

    def cancel(self):
        """Terminate any possible execution of the process running the func."""
        if self._process.is_alive():
            self._terminate_child_processes()
            self._process.terminate()

        # join process so the parent process can reap it - no zombies
        self._process.join(timeout=0.1)
        self._raise_exception()

    @property
    def ready(self):
        """Manage the status of "value" property."""
        if self._timeout < time.monotonic():
            self.cancel()
        return self._queue.full() and not self._queue.empty()

    @property
    def value(self):
        """Get data from the processing queue that was added by the func."""
        if self.ready is True:
            is_successful, ret_value = self._queue.get()
            if is_successful:
                return ret_value
            raise ret_value



# Copyright 2019-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Metrics registry to manage metrics for all configured relay clients.

The :class:`MetricsRegistry` class is the main client for which to create
and emit metrics. For example:

.. code-block:: python

    class MyTransform(KlioBaseTransform):
        def __init__(self):
            self.my_counter = self._klio.metrics.counter(
                name="foo",
                value=2,
                transform="mytransform",
                tags={"tag1": "value1"}
            )

        def process(self, element):
            # user code

            # immediately emits a metric
            self.my_counter.inc()

"""

import logging

from klio.metrics import dispatcher


class MetricsRegistry(object):
    """Main client to create and emit metrics.

    Args:
        relay_clients (list(klio.metrics.base.AbstractRelayClient)):
            configured relay clients.
    """

    def __init__(self, relay_clients, transform_name):
        self._relays = relay_clients
        self._transform_name = transform_name
        self._registry = {}

    def counter(self, name, value=0, **kwargs):
        """Get or create a counter.

        Creates a new counter if one is not found with a key of
        ``"counter_{name}_{transform}"`` of the given  ``transform``.

        New counters will be stored in memory for simple caching.

        Args:
            name (str): name of counter
            value (int): starting value of counter; defaults to 0
            kwargs (dict): keyword arguments passed to each configured
                relay clients' counter object.
        Returns:
            dispatcher.CounterDispatcher: instance of a counter dispatcher
        """
        transform_name = kwargs.pop("transform", self._transform_name)
        key = "counter_{}_{}".format(name, transform_name)
        if key in self._registry:
            return self._registry[key]

        counter = dispatcher.CounterDispatcher(
            relay_clients=self._relays,
            name=name,
            value=value,
            transform=transform_name,
            **kwargs
        )
        self._registry[key] = counter
        return counter

    def gauge(self, name, value=0, **kwargs):
        """Get or create a gauge.

        Creates a new gauge if one is not found with a key of
        "gauge_{name}_{transform}".

        New gauges will be stored in memory for simple caching.

        Args:
            name (str): name of gauge
            value (int): starting value of gauge; defaults to 0
            kwargs (dict): keyword arguments passed to each configured
                relay clients' gauge object.
        Returns:
            dispatcher.GaugeDispatcher: instance of a gauge dispatcher
        """
        transform_name = kwargs.pop("transform", self._transform_name)
        key = "gauge_{}_{}".format(name, transform_name)
        if key in self._registry:
            return self._registry[key]

        gauge = dispatcher.GaugeDispatcher(
            relay_clients=self._relays,
            name=name,
            value=value,
            transform=transform_name,
            **kwargs
        )
        self._registry[key] = gauge
        return gauge

    def timer(self, name, value=0, timer_unit="ns", **kwargs):
        """Get or create a timer.

        Creates a new timer if one is not found with a key of
        "timer_{name}_{transform}".

        New timers will be stored in memory for simple caching.

        Args:
            name (str): name of timer
            value (int): starting value of timer; defaults to 0
            timer_unit (str): desired unit of time; defaults to ns
            kwargs (dict): keyword arguments passed to each configured
                relay clients' timer object.
        Returns:
            dispatcher.TimerDispatcher: instance of a timer dispatcher
        """
        transform_name = kwargs.pop("transform", self._transform_name)
        key = "timer_{}_{}".format(name, transform_name)
        if key in self._registry:
            return self._registry[key]

        timer = dispatcher.TimerDispatcher(
            relay_clients=self._relays,
            name=name,
            value=value,
            transform=transform_name,
            timer_unit=timer_unit,
            **kwargs
        )
        self._registry[key] = timer
        return timer

    def marshal(self, metric):
        """Create a dictionary-representation of a given metric.

        Used when metric objects need to be pickled.

        Args:
            metric (dispatcher.BaseMetricDispatcher): metric instance to
                marshal.

        Returns:
            dict: the metric's data in dictionary form
        """
        data = {
            "type": metric.METRIC_TYPE,
            "name": metric.name,
            "value": metric.value,
        }
        data.update(metric.kwargs)
        return data

    def unmarshal(self, metric_data):
        """Create a metric instance based off of a dictionary.

        If "type" is not specified or is not one of "counter", "gauge",
        or "timer", it defaults to a gauge-type metric.

        Used when metrics objects need to be unpickled.

        Args:
            metric_data (dict): dictionary-representation of a given
                metric.

        Returns:
            dispatcher.BaseMetricDispatcher: a dispatcher relevant to
            metric type.
        """
        metric_type = metric_data.pop("type", None)
        name = metric_data.pop("name", None)
        value = metric_data.pop("value", None)
        kwargs = metric_data

        metric_method = {
            "counter": self.counter,
            "gauge": self.gauge,
            "timer": self.timer,
        }.get(metric_type)

        if not metric_method:
            msg = "Metric type '{}' not supported. Defaulting to gauge.".format(
                metric_type
            )
            logging.getLogger("klio.metrics").warning(msg)
            metric_method = self.gauge

        return metric_method(name=name, value=value, **kwargs)



# Copyright 2019-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging

import pytest

from klio_core import config


@pytest.fixture
def caplog(caplog):
    """Set global test logging levels."""
    caplog.set_level(logging.DEBUG)
    return caplog


@pytest.fixture
def job_config_dict():
    return {
        "metrics": {"logger": {}},
        "allow_non_klio_messages": False,
        "number_of_retries": 3,
        "inputs": [
            {
                "topic": "test-parent-job-out",
                "subscription": "test-parent-job-out-sub",
                "data_location": "gs://sigint-output/test-parent-job-out",
            }
        ],
        "outputs": [
            {
                "topic": "test-job-out",
                "data_location": "gs://sigint-output/test-job-out",
            }
        ],
        "dependencies": [
            {"gcp_project": "sigint", "job_name": "test-parent-job"}
        ],
        "more": "config",
        "that": {"the": "user"},
        "might": ["include"],
    }


@pytest.fixture
def pipeline_config_dict():
    return {
        "project": "test-project",
        "staging_location": "gs://some/stage",
        "temp_location": "gs://some/temp",
        "worker_harness_container_image": "gcr.io/sigint/foo",
        "streaming": True,
        "experiments": ["beam_fn_api"],
        "region": "us-central1",
        "num_workers": 3,
        "max_num_workers": 5,
        "disk_size_gb": 50,
        "worker_machine_type": "n1-standard-4",
        "runner": "direct",
        "autoscaling_algorithm": "THROUGHPUT_BASED",
        "update": False,
    }


@pytest.fixture
def config_dict(job_config_dict, pipeline_config_dict):
    return {
        "job_config": job_config_dict,
        "pipeline_options": pipeline_config_dict,
        "job_name": "test-job",
        "version": 1,
    }


@pytest.fixture
def klio_config(config_dict):
    return config.KlioConfig(config_dict)


@pytest.fixture
def mock_config(mocker, monkeypatch):
    mconfig = mocker.Mock(name="MockKlioConfig")
    mconfig.job_name = "a-job"
    mconfig.pipeline_options.streaming = True
    mconfig.pipeline_options.project = "not-a-real-project"

    mock_data_input = mocker.Mock(name="MockDataGcsInput")
    mock_data_input.type = "gcs"
    mock_data_input.location = "gs://hopefully-this-bucket-doesnt-exist"
    mock_data_input.file_suffix = ""
    mock_data_input.skip_klio_existence_check = True
    mconfig.job_config.data.inputs = [mock_data_input]
    monkeypatch.setattr(
        "klio.transforms.core.RunConfig.get", lambda: mconfig,
    )
    return mconfig



# Copyright 2019-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest

from googleapiclient import errors as gapi_errors

from klio.metrics import stackdriver as sd_metrics


@pytest.fixture
def mock_stackdriver_client(mocker, monkeypatch):
    mock_client = mocker.Mock()
    monkeypatch.setattr(
        sd_metrics.discovery, "build", lambda x, y: mock_client
    )
    return mock_client


@pytest.fixture
def client(klio_config, monkeypatch, mocker, mock_stackdriver_client):
    return sd_metrics.StackdriverLogMetricsClient(klio_config)


@pytest.fixture
def mock_discovery_build(mocker, monkeypatch):
    mock_discovery_build = mocker.Mock()
    monkeypatch.setattr(sd_metrics.discovery, "build", mock_discovery_build)
    return mock_discovery_build


@pytest.fixture
def counter():
    kwargs = {
        "name": "my-counter",
        "job_name": "test-job",
        "project": "test-project",
    }
    return sd_metrics.StackdriverLogMetricsCounter(**kwargs)


def test_stackdriver_client(client, mock_discovery_build):
    # sanity check
    assert not getattr(client._thread_local, "stackdriver_client", None)

    client._stackdriver_client

    assert getattr(client._thread_local, "stackdriver_client", None)
    mock_discovery_build.assert_called_once_with("logging", "v2")


@pytest.mark.parametrize("value", (None, 5))
def test_client_counter(client, value, mocker, monkeypatch, caplog):
    mock_init_metric = mocker.Mock()
    monkeypatch.setattr(
        sd_metrics.StackdriverLogMetricsCounter,
        "_init_metric",
        mock_init_metric,
    )

    kwargs = {}
    if value:
        kwargs["value"] = value

    counter = client.counter(name="my-counter", **kwargs)

    mock_init_metric.assert_called_once_with(client._stackdriver_client)
    assert isinstance(counter, sd_metrics.StackdriverLogMetricsCounter)

    if value:
        assert 1 == len(caplog.records)
        assert "WARNING" == caplog.records[0].levelname


def test_client_gauge(client, caplog):
    gauge = client.gauge(name="my-gauge")

    assert isinstance(gauge, sd_metrics.StackdriverLogMetricsGauge)
    assert 1 == len(caplog.records)
    assert "WARNING" == caplog.records[0].levelname


def test_client_timer(client, caplog):
    gauge = client.timer(name="my-timer")

    assert isinstance(gauge, sd_metrics.StackdriverLogMetricsTimer)
    assert 1 == len(caplog.records)
    assert "WARNING" == caplog.records[0].levelname


def test_counter_get_filter(counter):
    exp_filter = (
        'resource.type="dataflow_step" '
        'logName="projects/test-project/logs/'
        'dataflow.googleapis.com%2Fworker" '
        'jsonPayload.message:"[my-counter]"'
    )

    actual_filter = counter._get_filter()
    assert exp_filter == actual_filter


@pytest.mark.parametrize("raises", (False, True))
def test_counter_init_metric(raises, counter, mock_stackdriver_client, caplog):
    _mock_projects = mock_stackdriver_client.projects.return_value
    mock_req = _mock_projects.metrics.return_value.create
    if raises:
        mock_req.return_value.execute.side_effect = Exception("foo")

    counter._init_metric(mock_stackdriver_client)

    mock_req.assert_called_once_with(parent=counter.parent, body=counter.body)
    mock_req.return_value.execute.assert_called_once_with()

    if raises:
        assert 1 == len(caplog.records)
        assert "ERROR" == caplog.records[0].levelname


@pytest.mark.parametrize(
    "status,exp_log_level", ((409, "DEBUG"), (403, "ERROR"))
)
def test_counter_init_metric_api_error(
    status, exp_log_level, counter, mock_stackdriver_client, mocker, caplog
):
    mock_resp = mocker.Mock()
    mock_resp.status = status
    error = gapi_errors.HttpError(resp=mock_resp, content=b"foo")
    _mock_projects = mock_stackdriver_client.projects.return_value
    mock_req = _mock_projects.metrics.return_value.create
    mock_req.return_value.execute.side_effect = error

    counter._init_metric(mock_stackdriver_client)

    mock_req.assert_called_once_with(parent=counter.parent, body=counter.body)
    mock_req.return_value.execute.assert_called_once_with()

    assert 1 == len(caplog.records)
    assert exp_log_level == caplog.records[0].levelname



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys

import apache_beam as beam
import pytest

from apache_beam.options import pipeline_options
from apache_beam.testing import test_pipeline

from klio_core.proto import klio_pb2

from klio.transforms import helpers


IS_PY36 = sys.version_info < (3, 7)


class BaseTest:
    def get_current_job(self):
        job = klio_pb2.KlioJob()
        job.job_name = "a-job"
        job.gcp_project = "not-a-real-project"
        return job

    def get_other_job(self):
        job = klio_pb2.KlioJob()
        job.job_name = "other"
        job.gcp_project = "not-a-real-project"
        return job

    def assert_actual_is_expected(self, actual, expected):
        assert actual == expected, "Expected PColl was not processed"

    def assert_not_processed(self, *args, **kwargs):
        # We're not expecting to get called, so if it does fail, then we're
        # getting pcolls that we're not expecting. It may also mean
        # any call to assert_actual_is_expected was not called
        assert False, "Received an unexpected PColl"


# TODO: remove me when migration to v2 is done
@pytest.mark.usefixtures("mock_config")
class TestV1ExecutionMode(BaseTest):
    """v1 messages are processed or dropped depending on top-down or bottom up.
    """

    def prep_kmsg(self):
        msg = klio_pb2.KlioMessage()
        msg.data.entity_id = "s0m3_tr4ck_1d"
        return msg

    # unversioned incoming message
    def input_v0_kmsg(self):
        msg = self.prep_kmsg()
        return msg

    def input_v1_kmsg(self):
        msg = self.input_v0_kmsg()
        msg.version = klio_pb2.Version.V1
        return msg

    def exp_kmsg(self):
        msg = self.input_v1_kmsg()
        msg.data.element = b"s0m3_tr4ck_1d"
        return msg.SerializeToString()

    def test_top_down(self):
        pcoll = [
            self.input_v1_kmsg().SerializeToString(),
            self.input_v0_kmsg().SerializeToString(),
        ]
        exp_pcoll = self.exp_kmsg()

        with test_pipeline.TestPipeline() as p:
            in_pcol = p | beam.Create(pcoll)
            msg_version = in_pcol | helpers._KlioTagMessageVersion()
            should_process = msg_version.v1 | helpers._KlioV1CheckRecipients()

            _ = (
                should_process.process
                | "Assert expected processed"
                >> beam.Map(self.assert_actual_is_expected, expected=exp_pcoll)
            )
            _ = should_process.drop | "Assert no drops" >> beam.Map(
                self.assert_not_processed
            )
            _ = msg_version.v2 | "Assert no v2" >> beam.Map(
                self.assert_not_processed
            )

    def input_v0_kmsg_downstream(self):
        msg = self.input_v0_kmsg()
        msg.metadata.downstream.extend([self.get_current_job()])
        return msg

    def input_v1_kmsg_downstream(self):
        msg = self.input_v1_kmsg()
        msg.metadata.downstream.extend([self.get_current_job()])
        return msg

    def exp_kmsg_downstream(self):
        msg = self.input_v1_kmsg_downstream()
        msg.data.element = b"s0m3_tr4ck_1d"
        return msg.SerializeToString()

    def exp_v0_kmsg_downstream(self):
        msg = self.input_v1_kmsg_downstream()
        msg.data.element = b"s0m3_tr4ck_1d"
        msg.version = klio_pb2.Version.V1
        return msg.SerializeToString()

    def test_bottom_up(self):
        pcoll = [
            self.input_v1_kmsg_downstream().SerializeToString(),
            self.input_v0_kmsg_downstream().SerializeToString(),
        ]
        exp_pcoll = self.exp_kmsg_downstream()

        with test_pipeline.TestPipeline() as p:
            in_pcol = p | beam.Create(pcoll)
            msg_version = in_pcol | helpers._KlioTagMessageVersion()
            should_process = msg_version.v1 | helpers._KlioV1CheckRecipients()

            _ = (
                should_process.process
                | "Assert expected processed"
                >> beam.Map(self.assert_actual_is_expected, expected=exp_pcoll)
            )
            _ = should_process.drop | "Assert no dropped msgs" >> beam.Map(
                self.assert_not_processed
            )
            _ = msg_version.v2 | "Assert no v2 msgs" >> beam.Map(
                self.assert_not_processed
            )

    def input_v0_kmsg_drop(self):
        msg = self.input_v0_kmsg()
        msg.metadata.downstream.extend([self.get_other_job()])
        return msg

    def input_v1_kmsg_drop(self):
        msg = self.input_v1_kmsg()
        msg.metadata.downstream.extend([self.get_other_job()])
        return msg

    def exp_kmsg_drop(self):
        msg = self.input_v1_kmsg_drop()
        msg.data.element = b"s0m3_tr4ck_1d"
        return msg.SerializeToString()

    def test_bottom_up_drop(self):
        pcoll = [
            self.input_v1_kmsg_drop().SerializeToString(),
            self.input_v0_kmsg_drop().SerializeToString(),
        ]
        exp_pcoll = self.exp_kmsg_drop()

        with test_pipeline.TestPipeline() as p:
            in_pcol = p | beam.Create(pcoll)
            msg_version = in_pcol | helpers._KlioTagMessageVersion()
            should_process = msg_version.v1 | helpers._KlioV1CheckRecipients()

            _ = should_process.drop | "Assert expected dropped" >> beam.Map(
                self.assert_actual_is_expected, expected=exp_pcoll
            )
            _ = should_process.process | "Assert not processed" >> beam.Map(
                self.assert_not_processed
            )
            _ = msg_version.v2 | "Assert no v2" >> beam.Map(
                self.assert_not_processed
            )


@pytest.mark.usefixtures("mock_config")
class TestExecutionMode(BaseTest):
    """Messages are processed or dropped depending on top-down or bottom up.
    """

    def prep_kmsg(self):
        msg = klio_pb2.KlioMessage()
        msg.data.element = b"s0m3_tr4ck_1d"
        return msg

    # unversioned incoming message
    def input_v0_kmsg_anyone(self):
        msg = self.prep_kmsg()
        msg.metadata.intended_recipients.anyone.SetInParent()
        return msg

    def input_v2_kmsg_anyone(self):
        msg = self.input_v0_kmsg_anyone()
        msg.version = klio_pb2.Version.V2
        return msg

    def exp_kmsg_anyone(self):
        msg = self.input_v2_kmsg_anyone()
        return msg.SerializeToString()

    def test_top_down(self):
        pcoll = [
            self.input_v0_kmsg_anyone().SerializeToString(),
            self.input_v2_kmsg_anyone().SerializeToString(),
        ]
        exp_pcoll = self.exp_kmsg_anyone()

        with test_pipeline.TestPipeline() as p:
            in_pcol = p | beam.Create(pcoll)
            msg_version = in_pcol | helpers._KlioTagMessageVersion()
            should_process = msg_version.v2 | helpers.KlioCheckRecipients()

            _ = (
                should_process.process
                | "Assert expected processed"
                >> beam.Map(self.assert_actual_is_expected, expected=exp_pcoll)
            )
            _ = should_process.drop | "Assert no dropped msgs" >> beam.Map(
                self.assert_not_processed
            )
            _ = msg_version.v1 | "Assert no v1 msgs" >> beam.Map(
                self.assert_not_processed
            )

    def input_v0_kmsg_limited(self):
        msg = self.prep_kmsg()
        msg.metadata.intended_recipients.limited.recipients.extend(
            [self.get_current_job()]
        )
        return msg

    def input_v2_kmsg_limited(self):
        msg = self.input_v0_kmsg_limited()
        msg.version = klio_pb2.Version.V2
        return msg

    def exp_kmsg_limited(self):
        msg = self.input_v2_kmsg_limited()
        return msg.SerializeToString()

    def test_bottom_up(self):
        pcoll = [
            self.input_v0_kmsg_limited().SerializeToString(),
            self.input_v2_kmsg_limited().SerializeToString(),
        ]
        exp_pcoll = self.exp_kmsg_limited()

        with test_pipeline.TestPipeline() as p:
            in_pcol = p | beam.Create(pcoll)
            msg_version = in_pcol | helpers._KlioTagMessageVersion()
            should_process = msg_version.v2 | helpers.KlioCheckRecipients()

            _ = (
                should_process.process
                | "Assert expected processed"
                >> beam.Map(self.assert_actual_is_expected, expected=exp_pcoll)
            )
            _ = should_process.drop | "Assert no dropped msgs" >> beam.Map(
                self.assert_not_processed
            )
            _ = msg_version.v1 | "Assert no v1 msgs" >> beam.Map(
                self.assert_not_processed
            )

    def input_v0_kmsg_drop(self):
        msg = self.prep_kmsg()
        msg.metadata.intended_recipients.limited.recipients.extend(
            [self.get_other_job()]
        )
        return msg

    def input_v2_kmsg_drop(self):
        msg = self.input_v0_kmsg_drop()
        msg.version = klio_pb2.Version.V2
        return msg

    def exp_kmsg_drop(self):
        msg = self.input_v2_kmsg_drop()
        return msg.SerializeToString()

    def test_bottom_up_drop(self):
        pcoll = [
            self.input_v0_kmsg_drop().SerializeToString(),
            self.input_v2_kmsg_drop().SerializeToString(),
        ]
        exp_pcoll = self.exp_kmsg_drop()

        with test_pipeline.TestPipeline() as p:
            in_pcol = p | beam.Create(pcoll)
            msg_version = in_pcol | helpers._KlioTagMessageVersion()
            should_process = msg_version.v2 | helpers.KlioCheckRecipients()

            _ = should_process.drop | "Assert expected dropped" >> beam.Map(
                self.assert_actual_is_expected, expected=exp_pcoll
            )
            _ = (
                should_process.process
                | "Assert no processed msgs"
                >> beam.Map(self.assert_not_processed)
            )
            _ = msg_version.v1 | "Assert no v1 msgs" >> beam.Map(
                self.assert_not_processed
            )

    def in_v0_kmsg_trigger_children_of(self):
        msg = self.prep_kmsg()
        current_job = self.get_current_job()
        lmtd = msg.metadata.intended_recipients.limited
        lmtd.recipients.extend([current_job, self.get_other_job()])
        lmtd.trigger_children_of.job_name = current_job.job_name
        lmtd.trigger_children_of.gcp_project = current_job.gcp_project
        return msg

    def in_v2_kmsg_trigger_children_of(self):
        msg = self.in_v0_kmsg_trigger_children_of()
        msg.version = klio_pb2.Version.V2
        return msg

    def exp_kmsg_trigger_children_of(self):
        msg = self.prep_kmsg()
        msg.version = klio_pb2.Version.V2
        msg.metadata.intended_recipients.anyone.SetInParent()
        return msg.SerializeToString()

    def _prep_pipeline(self):
        pcoll = [
            self.in_v0_kmsg_trigger_children_of().SerializeToString(),
            self.in_v2_kmsg_trigger_children_of().SerializeToString(),
        ]
        pipeline = test_pipeline.TestPipeline()
        in_pcol = pipeline.apply(beam.Create(pcoll), pipeline)
        msg_version = in_pcol.apply(helpers._KlioTagMessageVersion())
        return msg_version

    def test_bottom_up_to_top_down(self):
        pcoll = [
            self.in_v0_kmsg_trigger_children_of().SerializeToString(),
            self.in_v2_kmsg_trigger_children_of().SerializeToString(),
        ]
        exp_pcoll = self.exp_kmsg_trigger_children_of()

        with test_pipeline.TestPipeline() as p:
            in_pcol = p | beam.Create(pcoll)
            msg_version = in_pcol | helpers._KlioTagMessageVersion()
            should_process = msg_version.v2 | helpers.KlioCheckRecipients()

            _ = (
                should_process.process
                | "Assert expected processed"
                >> beam.Map(self.assert_actual_is_expected, expected=exp_pcoll)
            )
            _ = should_process.drop | "Assert no dropped msgs" >> beam.Map(
                self.assert_not_processed
            )
            _ = msg_version.v1 | "Assert no v1 msgs" >> beam.Map(
                self.assert_not_processed
            )


def assert_audit(actual):
    job = klio_pb2.KlioJob()
    job.job_name = "a-job"
    job.gcp_project = "not-a-real-project"
    audit_log_item = klio_pb2.KlioJobAuditLogItem()
    audit_log_item.klio_job.CopyFrom(job)
    exp_msg = klio_pb2.KlioMessage()
    exp_msg.version = klio_pb2.Version.V2
    exp_msg.metadata.job_audit_log.extend([audit_log_item])
    expected = exp_msg.SerializeToString()

    assert expected == actual
    return actual


def test_update_klio_log(mocker, monkeypatch, caplog, mock_config):
    mock_ts = mocker.Mock()
    monkeypatch.setattr(klio_pb2.KlioJobAuditLogItem, "timestamp", mock_ts)

    kmsg = klio_pb2.KlioMessage()
    kmsg.version = klio_pb2.Version.V2
    assert not kmsg.metadata.job_audit_log  # sanity check

    with test_pipeline.TestPipeline() as p:
        in_pcol = p | beam.Create([kmsg.SerializeToString()])
        act_pcol = in_pcol | helpers.KlioUpdateAuditLog()
        _ = act_pcol | beam.Map(assert_audit)

    exp_log = (
        "KlioMessage full audit log - Entity ID:  - Path: not-a-real-project::"
        "a-job (current job)"
    )
    for rec in caplog.records:
        if exp_log in rec.message:
            assert True
            break
    else:
        assert False, "Expected debug audit log not found"


@pytest.mark.skipif(IS_PY36, reason="This test fails to pickle on 3.6")
def test_trigger_upstream_job(mock_config, mocker, capsys):
    mock_gcs_client = mocker.patch("klio.transforms._helpers.gcsio.GcsIO")
    mock_gcs_client.return_value.exists.return_value = False
    mock_pubsub_client = mocker.patch("google.cloud.pubsub.PublisherClient")

    kmsg = klio_pb2.KlioMessage()
    kmsg.data.element = b"does_not_exist"

    exp_current_job = klio_pb2.KlioJob()
    exp_current_job.job_name = "a-job"
    exp_current_job.gcp_project = "not-a-real-project"
    exp_upstream_job = klio_pb2.KlioJob()
    exp_upstream_job.job_name = "upstream-job"
    exp_upstream_job.gcp_project = "upstream-project"
    exp_kmsg = klio_pb2.KlioMessage()
    exp_kmsg.version = klio_pb2.Version.V2
    exp_kmsg.data.element = b"does_not_exist"
    exp_lmtd = exp_kmsg.metadata.intended_recipients.limited
    exp_lmtd.recipients.extend([exp_upstream_job, exp_current_job])
    exp_lmtd.trigger_children_of.CopyFrom(exp_current_job)

    options = pipeline_options.PipelineOptions([])
    options.view_as(pipeline_options.StandardOptions).streaming = True

    with test_pipeline.TestPipeline(options=options) as p:
        in_pcol = p | beam.Create([kmsg.SerializeToString()])
        input_data = in_pcol | helpers.KlioGcsCheckInputExists()

        _ = input_data.not_found | helpers.KlioTriggerUpstream(
            upstream_job_name="upstream-job",
            upstream_topic="projects/upstream-project/topics/does-not-exist",
        )

    mock_gcs_client.return_value.exists.assert_called_once_with(
        "gs://hopefully-this-bucket-doesnt-exist/does_not_exist"
    )
    mock_pubsub_client.return_value.publish.assert_called_once_with(
        mock_pubsub_client.return_value.topic_path.return_value,
        exp_kmsg.SerializeToString(),
    )



# Copyright 2019-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging
import threading

import pytest

from klio.metrics import logger


@pytest.fixture
def client(klio_config):
    return logger.MetricsLoggerClient(klio_config)


@pytest.fixture
def metric():
    return logger.LoggerCounter(name="fixture-counter")


@pytest.mark.parametrize(
    "metrics_config,exp_log_level",
    (
        ({}, logging.DEBUG),
        ({"logger": True}, logging.DEBUG),
        ({"logger": {"key": "value"}}, logging.DEBUG),
        ({"logger": {"level": "info"}}, logging.INFO),
        ({"logger": {"level": "INFO"}}, logging.INFO),
        ({"logger": {"level": "notalevel"}}, logging.DEBUG),
    ),
)
def test_client_set_log_level(metrics_config, exp_log_level, klio_config):
    klio_config.job_config.metrics = metrics_config
    client = logger.MetricsLoggerClient(klio_config)

    assert exp_log_level == client.log_level


@pytest.mark.parametrize(
    "metrics_config,exp_timer_unit",
    (
        ({}, "ns"),
        ({"logger": True}, "ns"),
        ({"logger": {"timer_unit": "seconds"}}, "s"),
        ({"logger": {"timer_unit": "notaunit"}}, "ns"),
    ),
)
def test_client_set_timer_unit(metrics_config, exp_timer_unit, klio_config):
    klio_config.job_config.metrics = metrics_config
    client = logger.MetricsLoggerClient(klio_config)

    assert exp_timer_unit == client.timer_unit


def test_client_logger(client):
    ret_logger = client.logger

    assert logging.getLogger("klio.metrics") == ret_logger


def test_client_logger_local(client, monkeypatch):
    logger_patch = logging.getLogger("klio.metrics.patch")
    thread_local_patch = threading.local()
    thread_local_patch.klio_metrics_logger = logger_patch
    monkeypatch.setattr(client, "_thread_local", thread_local_patch)

    ret_logger = client.logger

    assert logger_patch == ret_logger


def test_client_unmarshal(client, metric):
    exp_metric_dict = {
        "name": "fixture-counter",
        "value": 0,
        "transform": None,
        "tags": {"metric_type": "counter"},
    }
    unmarshalled = client.unmarshal(metric)
    assert exp_metric_dict == unmarshalled


def test_client_emit(client, metric, caplog):
    exp_log_record = (
        "[fixture-counter] value: 0 transform: 'None' "
        "tags: {'metric_type': 'counter'}"
    )
    client.emit(metric)

    assert 1 == len(caplog.records)
    assert exp_log_record == caplog.records[0].message


def test_client_counter(client):
    counter = client.counter(name="my-counter")
    assert isinstance(counter, logger.LoggerCounter)


def test_client_gauge(client):
    gauge = client.gauge(name="my-gauge")
    assert isinstance(gauge, logger.LoggerGauge)


@pytest.mark.parametrize(
    "timer_unit,config_timer_unit,exp_timer_unit",
    (
        (None, "ns", "ns"),
        (None, None, "ns"),
        ("s", None, "s"),
        ("microseconds", None, "us"),
        ("notaunit", None, "ns"),
    ),
)
def test_client_timer(
    timer_unit, config_timer_unit, exp_timer_unit, client, monkeypatch
):
    if config_timer_unit:
        monkeypatch.setattr(client, "timer_unit", config_timer_unit)
    timer = client.timer(name="my-timer", timer_unit=timer_unit)
    assert isinstance(timer, logger.LoggerTimer)
    assert exp_timer_unit == timer.tags["unit"]


def test_logger_counter():
    expected_tags = {"metric_type": "counter"}

    counter = logger.LoggerCounter(name="my-counter")
    assert expected_tags == counter.tags


def test_logger_gauge():
    expected_tags = {"metric_type": "gauge"}

    gauge = logger.LoggerGauge(name="my-gauge")
    assert expected_tags == gauge.tags


def test_logger_timer():
    expected_tags = {"metric_type": "timer", "unit": "ns"}
    timer = logger.LoggerTimer(name="my-timer")
    assert expected_tags == timer.tags

    expected_tags["unit"] = "s"
    timer = logger.LoggerTimer(name="my-timer", timer_unit="s")
    assert expected_tags == timer.tags


# pseudo-integration test
@pytest.mark.parametrize("disabled,exp", ((True, 0), (False, 1)))
def test_logger_disabled(disabled, exp, client, metric, caplog, monkeypatch):
    caplog.set_level(logging.DEBUG, logger="klio.metrics")
    # sanity check / clear out thread local
    client._thread_local.klio_metrics_logger = None

    monkeypatch.setattr(client, "disabled", disabled)

    client.emit(metric)

    assert exp == len(caplog.records)



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""For internal use only; no backwards-compatibility guarantees."""

import enum
import functools
import warnings


class AnnotatedStates(enum.Enum):
    DEPRECATED = "deprecated"
    EXPERIMENTAL = "experimental"


# adapted from https://github.com/apache/beam/blob/9c3941fc/
#                      sdks/python/apache_beam/utils/annotations.py
class KlioDeprecationWarning(DeprecationWarning):
    """Klio-specific deprecation warnings."""


class KlioFutureWarning(FutureWarning):
    """Klio-specific deprecation warnings."""


# Don't ignore klio deprecation warnings! (future warnings ok)
warnings.simplefilter("once", KlioDeprecationWarning)


def is_original_process_func(clsdict, bases, base_class=None):
    """Only wrap the original `process` function.

    Without these (minimal) checks, the `process` function would be
    wrapped at least twice (the original `process` function from the
    user's DoFn, and our wrapped/decorated one), essentially causing
    any call to `process` (and the decorator) to be called at least
    twice.

    Args:
        clsdict (dict): dictionary of items for the class being
            instantiated.
        bases (tuple(class)): base class(es) of the class being
            instantiated.
    Returns:
        (bool) whether or not to wrap the `process` method of the class
            being instantiated.
    """
    if "process" not in clsdict:
        return False
    # ignore classes that don't inherit from our base class
    base_cls_names = [b.__name__ for b in bases]
    if base_class and base_class not in base_cls_names:
        return False
    # if the value of clsdict["process"] is not a meth/func
    if not callable(clsdict["process"]):
        return False
    # if the value of clsdict["process"] is already "new_process"
    if getattr(clsdict["process"], "__name__") != "process":
        return False
    return True


# adapted from https://github.com/apache/beam/blob/9c3941fc/
#                      sdks/python/apache_beam/utils/annotations.py
def annotate(state, since=None, current=None, message=None):
    """Decorates an API with a `deprecated` or `experimental` annotation.

    When a user uses a objected decorated with this annotation, they
    will see a `KlioFutureWarning` or `KlioDeprecationWarning` during
    runtime.

    Args:
        state (AnnotatedStates): the kind of annotation (AnnotatedStates.
            DEPRECATED or AnnotatedStates.EXPERIMENTAL).
        since: the version that causes the annotation (used for
            AnnotatedStates.DEPRECATED when no `message` is given;
            ignored for AnnotatedStates.EXPERIMENTAL).
        current: the suggested replacement function.
        message: if the default message does not suffice, the message
            can be changed using this argument. Default message for

    Returns:
        The decorator for the API.
    """

    def wrapper(func):
        @functools.wraps(func)
        def inner(*args, **kwargs):
            warning_type = KlioFutureWarning
            if state == AnnotatedStates.DEPRECATED:
                warning_type = KlioDeprecationWarning

            warn_message = message
            if message is None:
                addl_ctx = (
                    " and is subject to incompatible changes, or removal "
                    "in a future release of Klio."
                )
                if state == AnnotatedStates.DEPRECATED:
                    _since = " since {}".format(since) if since else ""
                    _current = (
                        ". Use {} instead".format(current) if current else ""
                    )
                    addl_ctx = "{}{}.".format(_since, _current)

                msg_kwargs = {
                    "obj": func.__name__,
                    "annotation": state.value,
                    "addl_ctx": addl_ctx,
                }
                warn_message = "'{obj}' is {annotation}{addl_ctx}".format(
                    **msg_kwargs
                )

            warnings.warn(warn_message, warning_type, stacklevel=2)
            return func(*args, **kwargs)

        return inner

    return wrapper


# partials for ease of use
deprecated = functools.partial(annotate, state=AnnotatedStates.DEPRECATED)
experimental = functools.partial(
    annotate, state=AnnotatedStates.EXPERIMENTAL, since=None
)



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json
import os

import apache_beam as beam
from apache_beam.io import avroio as beam_avroio
from apache_beam.io.gcp import bigquery as beam_bq
from fastavro import parse_schema

from klio_core.proto import klio_pb2


class BaseKlioIOException(Exception):
    """Base IO exception."""


class KlioMissingConfiguration(BaseKlioIOException):
    """Required configuration is missing."""


class _KlioReadWrapper(beam.io.Read):
    """Klio-ified `beam.io.Read` class.

    Handle the case if a user invokes an already-wrapped Klio IO
    transform (via `_KlioTransformMixin`) with `beam.io.Read`, as this
    is not possible to double-wrap.
    """

    def __init__(self, *args, **kwargs):
        if not len(args):
            # I would think we'd never get here, but this is just in case
            # something real funky happens, like this semi-private class
            # is not instantiated with an IO transform
            super(_KlioReadWrapper, self).__init__(*args, **kwargs)

        self.__wrapped_transform = args[0].__class__.__name__
        super(_KlioReadWrapper, self).__init__(*args, **kwargs)

    def is_bounded(self, *args, **kwargs):
        # beam.io.Read invokes beam.io.Read.source.is_bounded
        # (beam.io.Read.source == self/this instance) and throw an
        # AttributeError if is_bounded not defined. Here we define
        # the method but raise our own expection telling the user not to
        # invoke an alread-wrapped Klio IO Transform with `beam.io.Read()`
        raise TypeError(
            "Error reading from `{}`. If the transform was wrapped with "
            "`apache_beam.io.Read`, remove that wrapper and try again.".format(
                self.__wrapped_transform
            )
        )


class _KlioWrapIOMetaclass(type):
    def __call__(self, *args, **kwargs):
        # Some IO transforms require invocation to be wrapped with
        # `beam.io.Read()`, and others do not. This allows for the API
        # to be the same no matter what by wrapping `beam.io.Read` if
        # the transform requires it. If a user wraps a Klio IO transform
        # again (i.e. `beam.io.Read(_AlreadyWrappedKlioTransform))`), a
        # human-friendly `TypeError` will raise (see
        # `_KlioReadWrapper.is_bounded`)
        if self._REQUIRES_IO_READ_WRAP:
            return _KlioReadWrapper(
                super(_KlioWrapIOMetaclass, self).__call__(*args, **kwargs)
            )
        return super(_KlioWrapIOMetaclass, self).__call__(*args, **kwargs)


class _KlioTransformMixin(metaclass=_KlioWrapIOMetaclass):
    """Common properties for klio v2 IO transforms."""

    # whether or not the transform needs to be invoked by beam.io.Read()
    _REQUIRES_IO_READ_WRAP = False


class _KlioReadFromTextSource(beam.io.textio._TextSource):
    """Parses a text file as newline-delimited elements.
       Supports newline delimiters '\n' and '\r\n

    Returns:
        (str) KlioMessage serialized as a string
    """

    def read_records(self, file_name, range_tracker):
        records = super(_KlioReadFromTextSource, self).read_records(
            file_name, range_tracker
        )

        for record in records:
            record_as_bytes = record.encode("utf-8")
            message = klio_pb2.KlioMessage()
            message.version = klio_pb2.Version.V2
            message.metadata.intended_recipients.anyone.SetInParent()
            message.data.element = record_as_bytes
            yield message.SerializeToString()


class KlioReadFromText(beam.io.ReadFromText, _KlioTransformMixin):
    """Read from a local or GCS file with each new line as a
    ``KlioMessage.data.element``.
    """

    _source_class = _KlioReadFromTextSource


class _KlioReadFromBigQueryMapper(object):
    """Wrapper class to provide a ``beam.Map`` object that converts a row of a
    ``ReadFromBigQuery`` to a properly formatted ``KlioMessage``.
    """

    def __init__(self, klio_message_columns=None):
        self.__klio_message_columns = klio_message_columns

    def _generate_klio_message(self):
        message = klio_pb2.KlioMessage()
        message.version = klio_pb2.Version.V2
        message.metadata.intended_recipients.anyone.SetInParent()

        # TODO: this is where we should add (relevant) KlioMessage.metadata;
        # (1) One thing to figure out is the klio_pb2.KlioJob definition,
        # particularly the JobInput definition, in light of KlioConfig v2.
        # Once that's figured out, we should at least populate the
        # job audit log.
        # (2) Another thing to figure out is force/ping. In streaming, messages
        # are individually marked as force or ping when needed. However,
        # users aren't able to tag individual messages generated from a row
        # of BQ data as force/ping, and it's probably very difficult for us
        # to provide a way to do that. So, should we allow users to at least
        # globally set force/ping on their event input config in klio-job.yaml?
        # Potentially.
        return message

    def _map_row_element(self, row):
        # NOTE: this assumes that the coder being used (default is
        # beam.io.gcp.bigquery_tools.RowAsDictJsonCoder, otherwise set in
        # klio-job.yaml) is JSON serializable (since the default is just
        # a plain dictionary). This assumption might break if someone
        # provides a different coder.
        # NOTE: We need to have the row elements be bytes, so if it is
        # a dictionary, we json.dumps into a str to convert to bytes,
        # but that may need to change if we want to support other coders
        data = {}
        if self.__klio_message_columns:
            if len(self.__klio_message_columns) == 1:
                data = row[self.__klio_message_columns[0]]

            else:
                for key, value in row.items():
                    if key in self.__klio_message_columns:
                        data[key] = value
                data = json.dumps(data)

        else:
            data = json.dumps(row)
        return data

    def _map_row(self, row):
        message = self._generate_klio_message()
        message.data.element = bytes(self._map_row_element(row), "utf-8")
        return message.SerializeToString()

    def as_beam_map(self):
        return "Convert to KlioMessage" >> beam.Map(self._map_row)


# Note: copy-pasting the docstrings of `ReadFromBigQuery` so that we can
# include our added parameter (`klio_message_columns`) in the API
# documentation (via autodoc). If we don't do this, then just the parent
# documentation will be shown, excluding our new parameter.
class KlioReadFromBigQuery(beam.PTransform, _KlioTransformMixin):
    """Read data from BigQuery.

    This PTransform uses a BigQuery export job to take a snapshot of the table
    on GCS, and then reads from each produced file. File format is Avro by
    default.

    Args:
        table (str, callable, ValueProvider): The ID of the table, or a callable
            that returns it. The ID must contain only letters ``a-z``, ``A-Z``,
            numbers ``0-9``, or underscores ``_``. If dataset argument is
            :data:`None` then the table argument must contain the entire table
            reference specified as: ``'DATASET.TABLE'``
            or ``'PROJECT:DATASET.TABLE'``.
            If it's a callable, it must receive one argument representing an
            element to be written to BigQuery, and return
            a TableReference, or a string table name as specified above.
        dataset (str): The ID of the dataset containing this table or
            :data:`None` if the table reference is specified entirely by
            the table argument.
        project (str): The ID of the project containing this table.
        klio_message_columns (list): A list of fields (``str``) that should
            be assigned to ``KlioMessage.data.element``.

            .. note::

                If more than one field is provided, the results including the
                column names will be serialized to JSON before assigning to
                ``KlioMessage.data.element``. (e.g. ``'{"field1": "foo",
                "field2": bar"}'``). If only one field is provided, just the
                value will be assigned to ``KlioMessage.data.element``.

        query (str, ValueProvider): A query to be used instead of arguments
            table, dataset, and project.
        validate (bool): If :data:`True`, various checks will be done when
            source gets initialized (e.g., is table present?).
            This should be :data:`True` for most scenarios
            in order to catch errors as early as possible
            (pipeline construction instead of pipeline execution).
            It should be :data:`False` if the table is created during pipeline
            execution by a previous step.
        coder (~apache_beam.coders.coders.Coder): The coder for the table
            rows. If :data:`None`, then the default coder is
            _JsonToDictCoder, which will interpret every row as a JSON
            serialized dictionary.
        use_standard_sql (bool): Specifies whether to use BigQuery's standard
            SQL dialect for this query. The default value is :data:`False`.
            If set to :data:`True`, the query will use BigQuery's updated SQL
            dialect with improved standards compliance.
            This parameter is ignored for table inputs.
        flatten_results (bool): Flattens all nested and repeated fields in the
            query results. The default value is :data:`True`.
        kms_key (str): Optional Cloud KMS key name for use when creating new
            temporary tables.
        gcs_location (str, ValueProvider): The name of the Google Cloud Storage
            bucket where the extracted table should be written as a string or
            a :class:`~apache_beam.options.value_provider.ValueProvider`. If
            :data:`None`, then the temp_location parameter is used.
        bigquery_job_labels (dict): A dictionary with string labels to be passed
            to BigQuery export and query jobs created by this transform. See:
            https://cloud.google.com/bigquery/docs/reference/rest/v2/\
Job#JobConfiguration
        use_json_exports (bool): By default, this transform works by exporting
            BigQuery data into Avro files, and reading those files. With this
            parameter, the transform will instead export to JSON files.
            JSON files are slower to read due to their larger size.
            When using JSON exports,
            the BigQuery types for DATE, DATETIME, TIME, and TIMESTAMP will be
            exported as strings.

            This behavior is consistent with BigQuerySource.
            When using Avro exports,
            these fields will be exported as native Python
            types (datetime.date, datetime.datetime, datetime.datetime,
            and datetime.datetime respectively). Avro exports are recommended.
            To learn more about BigQuery types, and Time-related type
            representations,
            see:
            https://cloud.google.com/bigquery/docs/reference/standard-sql/\
data-types
            To learn more about type conversions between BigQuery and Avro, see:
            https://cloud.google.com/bigquery/docs/loading-data-cloud-\
storage-avro#avro_conversions
     """

    def __init__(self, *args, klio_message_columns=None, **kwargs):
        self.__reader = beam_bq.ReadFromBigQuery(*args, **kwargs)
        self.__mapper = _KlioReadFromBigQueryMapper(klio_message_columns)

    def expand(self, pcoll):
        return pcoll | self.__reader | self.__mapper.as_beam_map()


class KlioWriteToBigQuery(beam.io.WriteToBigQuery, _KlioTransformMixin):
    """Writes to BigQuery table with each row as ``KlioMessage.data.element``.
    """

    # Note: Not using BigQuerySink due to it only being available for
    # batch. See https://beam.apache.org/releases/pydoc/2.22.0/
    # apache_beam.io.gcp.bigquery.html?highlight=bigquerysink
    # #apache_beam.io.gcp.bigquery.BigQuerySink

    _REQUIRES_IO_READ_WRAP = False

    def __unwrap(self, encoded_element):
        message = klio_pb2.KlioMessage()
        message.ParseFromString(encoded_element)
        data = json.loads(message.data.payload)

        return data

    def expand(self, pcoll):
        return super().expand(pcoll | beam.Map(self.__unwrap))


class _KlioTextSink(beam.io.textio._TextSink):
    """A :class:`~apache_beam.transforms.ptransform.PTransform`
       for writing to text files. Takes a PCollection of KlioMessages
       and writes the elements to a textfile
    """

    def write_record(self, file_handle, encoded_element):
        """Writes a single encoded record.
        Args:
            file_handle (str): a referential identifier that points to an
                audio file found in the configured output data location.
            encoded_element (KlioMessage): KlioMessage
        """
        message = klio_pb2.KlioMessage()
        message.ParseFromString(encoded_element)
        record = message.data.element
        super(_KlioTextSink, self).write_encoded_record(file_handle, record)


class KlioWriteToText(beam.io.textio.WriteToText):
    """Write to a local or GCS file with each new line as
    ``KlioMessage.data.element``.
    """

    def __init__(self, *args, **kwargs):
        self._sink = _KlioTextSink(*args, **kwargs)


# note: fast avro is default for py3 on beam
class _KlioFastAvroSource(beam_avroio._FastAvroSource):
    def read_records(self, file_name, range_tracker):
        records = super(_KlioFastAvroSource, self).read_records(
            file_name=file_name, range_tracker=range_tracker
        )
        for record in records:
            message = klio_pb2.KlioMessage()
            message.version = klio_pb2.Version.V2
            message.metadata.intended_recipients.anyone.SetInParent()
            # If an element is sent then we set the element
            # to handle event reading
            # If "element" is not present then we stuff the record
            # into the message element
            message.data.element = (
                record["element"]
                if "element" in record
                else bytes(json.dumps(record).encode("utf-8"))
            )
            yield message.SerializeToString()


# define an I/O transform using the klio-specific avro source
# note: fast avro is default for py3 on beam
# Note: copy-pasting the docstrings of `ReadFromAvro` so that we can
# include our added parameter (`location`) in the API
# documentation (via autodoc) and drop `use_fastavro` since we default to
# True. If we don't do this, then just the parent documentation will be shown,
# excluding our new parameter and including an unavailable parameter
# (`location` and `use_fastavro` respectively)
class KlioReadFromAvro(beam.io.ReadFromAvro):
    """Read avro from a local directory or GCS bucket.

    Data from avro is dumped into JSON and assigned to ``KlioMessage.data.
    element``.

    ``KlioReadFromAvro`` is the default read for event input config type avro.
    However, ``KlioReadFromAvro`` can also be called explicity in a pipeline.

    Example pipeline reading in elements from an avro file:

    .. code-block:: python

        def run(pipeline, config):
            initial_data_path = os.path.join(DIRECTORY_TO_AVRO, "twitter.avro")
            pipeline | io_transforms.KlioReadFromAvro(
                    file_pattern=initial_data_path
            ) | beam.ParDo(transforms.HelloKlio())

    Args:
      file_pattern (str): the file glob to read.
      location (str): local or GCS path of file(s) to read.
      min_bundle_size (int): the minimum size in bytes, to be considered when
        splitting the input into bundles.
      validate (bool): flag to verify that the files exist during the pipeline
        creation time.
    """

    _REQUIRES_IO_READ_WRAP = True

    def __init__(
        self,
        file_pattern=None,
        location=None,
        min_bundle_size=0,
        validate=True,
    ):
        file_pattern = self._get_file_pattern(file_pattern, location)

        super(KlioReadFromAvro, self).__init__(
            file_pattern=file_pattern,
            min_bundle_size=min_bundle_size,
            validate=validate,
            use_fastavro=True,
        )

        self._source = _KlioFastAvroSource(
            file_pattern, min_bundle_size, validate=validate
        )

    def _get_file_pattern(self, file_pattern, location):
        # TODO: this should be a validator in klio_core.config
        if not any([file_pattern, location]):
            raise KlioMissingConfiguration(
                "Must configure at least one of the following keys when "
                "reading from avro: `file_pattern`, `location`."
            )

        if all([file_pattern, location]):
            file_pattern = os.path.join(location, file_pattern)

        elif file_pattern is None:
            file_pattern = location

        return file_pattern


# note: fast avro is default for py3 on beam
class _KlioFastAvroSink(beam_avroio._FastAvroSink):
    def write_record(self, writer, encoded_element):
        message = klio_pb2.KlioMessage()
        message.ParseFromString(encoded_element)
        record = {"element": message.data.element}
        super(_KlioFastAvroSink, self).write_record(
            writer=writer, value=record
        )


# Note of caution: In the past problems have arisen due to
# changes to internal beam classes.
# If this occurs, consider writing this as a custom PTransform
# that bundles a ``beam.Map`` with the standard ``WriteToAvro``
# Refer to ``KlioReadFromBigQuery`` as an example.
class KlioWriteToAvro(beam.io.WriteToAvro):
    """Write avro to a local directory or GCS bucket.

    ``KlioMessage.data.element`` data is parsed out
    and dumped into avro format.

    ``KlioWriteToAvro`` is the default write for event output config type avro.
    However, ``KlioWriteToAvro`` can also be called explicity in a pipeline.

    Example pipeline for writing elements to an avro file:

    .. code-block:: python

        def run(input_pcol, config):
            output_gcs_location = "gs://test-gcs-location"
            return (
                input_pcol
                | beam.ParDo(HelloKlio())
                | transforms.io.KlioWriteToAvro(location=output_gcs_location)
            )

    Args:
      file_path_prefix (str): The file path to write to
      location (str): local or GCS path to write to
      schema (str): The schema to use, as returned by avro.schema.parse
      codec (str): The codec to use for block-level compression.
            defaults to 'deflate'
      file_name_suffix (str): Suffix for the files written.
      num_shards (int): The number of files (shards) used for output.
      shard_name_template (str): template string for shard number and count
      mime_type (str): The MIME type to use for the produced files.
        Defaults to "application/x-avro"
    """

    KLIO_SCHEMA_OBJ = {
        "namespace": "klio.avro",
        "type": "record",
        "name": "KlioMessage",
        "fields": [{"name": "element", "type": "bytes"}],
    }

    def __init__(
        self,
        file_path_prefix=None,
        location=None,
        schema=parse_schema(KLIO_SCHEMA_OBJ),
        codec="deflate",
        file_name_suffix="",
        num_shards=0,
        shard_name_template=None,
        mime_type="application/x-avro",
    ):

        file_path = self._get_file_path(file_path_prefix, location)

        super(KlioWriteToAvro, self).__init__(
            file_path_prefix=file_path,
            schema=schema,
            codec=codec,
            file_name_suffix=file_name_suffix,
            num_shards=num_shards,
            shard_name_template=shard_name_template,
            mime_type=mime_type,
            use_fastavro=True,
        )

        self._sink = _KlioFastAvroSink(
            file_path,
            schema,
            codec,
            file_name_suffix,
            num_shards,
            shard_name_template,
            mime_type,
        )

    def _get_file_path(self, file_path_prefix, location):
        # TODO: this should be a validator in klio_core.config
        if not any([file_path_prefix, location]):
            raise KlioMissingConfiguration(
                "Must configure at least one of the following keys when "
                "writing to avro: `file_path_prefix`, `location`."
            )

        if all([file_path_prefix, location]):
            file_path_prefix = os.path.join(location, file_path_prefix)

        elif file_path_prefix is None:
            file_path_prefix = location

        return file_path_prefix



# Copyright 2019-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest
import six

from klio.metrics import base as base_metrics


class FakeObject(object):
    pass


@pytest.mark.parametrize("obj", [None, FakeObject()])
def test_abstract_attr(obj):
    ret_obj = base_metrics.abstract_attr(obj)

    if obj:
        assert isinstance(ret_obj, FakeObject)
    else:
        assert isinstance(ret_obj, base_metrics._DummyAttribute)

    assert hasattr(ret_obj, "__isabstractattr__")


def test__has_abstract_attributes_implemented_decorator():
    class TestMeta(six.with_metaclass(base_metrics._ABCBaseMeta)):
        @property
        @base_metrics.abstract_attr
        def foo(self):
            pass

    class TestABC(TestMeta):
        foo = "bar"

    rc = TestABC()
    assert hasattr(rc, "foo")


def test_has_abstract_attributes_implemented_cls_attr():
    class TestMeta(six.with_metaclass(base_metrics._ABCBaseMeta)):
        foo = base_metrics.abstract_attr()

    class TestABC(TestMeta):
        foo = "bar"

    rc = TestABC()
    assert hasattr(rc, "foo")


def test_has_abstract_attributes_implemented_raises():
    class TestMeta(six.with_metaclass(base_metrics._ABCBaseMeta)):
        foo = base_metrics.abstract_attr()

    with pytest.raises(NotImplementedError):

        class TestABC(TestMeta):
            not_foo = "nope"


def test_abstract_relay_client():
    class RelayClient(base_metrics.AbstractRelayClient):
        RELAY_CLIENT_NAME = "test-relay-client"

        def unmarshal(self, metric):
            return {}

        def emit(self, metric):
            return {}

        def counter(self, name, value=0, transform=None):
            return {}

        def gauge(self, name, value=0, transform=None):
            return {}

        def timer(self, name, value=0, transform=None):
            return {}

    rc = RelayClient({})
    assert hasattr(rc, "RELAY_CLIENT_NAME")
    assert hasattr(rc, "unmarshal")
    assert hasattr(rc, "emit")
    assert hasattr(rc, "counter")
    assert hasattr(rc, "gauge")
    assert hasattr(rc, "timer")


def test_abstract_relay_client_raises():
    with pytest.raises(NotImplementedError):

        class RelayClient(base_metrics.AbstractRelayClient):
            pass


def test_base_metric():
    class MyMetric(base_metrics.BaseMetric):
        def __init__(self, name, value=0, transform=None, tags=None, **kwargs):
            super(MyMetric, self).__init__(
                name, value=value, transform=transform
            )
            self.tags = tags

    my_metric_inst = MyMetric(name="foo")

    assert isinstance(my_metric_inst, base_metrics.BaseMetric)
    assert hasattr(my_metric_inst, "name")
    assert hasattr(my_metric_inst, "value")
    assert hasattr(my_metric_inst, "transform")
    assert hasattr(my_metric_inst, "update")

    assert 0 == my_metric_inst.value
    my_metric_inst.update(1)
    assert 1 == my_metric_inst.value
